--[[
  Guild Event Manager (GEM) - Version 3
  by Kiki from European "Conseil des Ombres" (Horde)
  ----------------------------------------
  Blizzard Calendar API module
]]

--[[
  Exported API Functions:


]]

--[[
 TODO:
  - Quand on fait un AddExternal, faire un "InvitePlayer" dans le bliz cal
  - Gerer la deadline: Locker l'event des que la deadline est passee
  - Si on desactive la synchro des events avec blizCal, virer tous les events qui sont a moi (de meme qd on active, creer mes events dans le cal)
  - Promote les assistants GEM, dans le cal
  
]]

--[[

  -- Actions
  CalendarEventAvailable() - Accepts the inviation to the currently open event 
  CalendarEventDecline() - Declines the invitation to the currently open event

  -- Setup
  CalendarNewEvent() - Creates and selected a new event 
  CalendarNewGuildEvent(minLevel, maxLevel, minRank) - Replaces the invite list of the selected new event with the specified guild members  
  CalendarEventSetDate(month, day, year) 
  CalendarEventSetDescription(description)  
  CalendarEventSetSize - ??
  CalendarEventSetTime(hour, minute) 
  CalendarEventSetTitle(title)
  CalendarEventSetType(type) -> one of CALENDAR_EVENTTYPE_xx
  CalendarEventSetTextureID(text_id) -> avec le mode donjon ou raid

  -- Control
  CalendarCloseEvent() - Closes the selected event without saving it  
  CalendarAddEvent() - Saves the selected event (new events only, requires hardware input to call)
  CalendarUpdateEvent() - Saves the selected event (existing events only, requires hardware input to call)
  CalendarEventInvite("Player") - Invite player to currently selected event
  CalendarEventRemoveInvite(inviteIndex) 
  CalendarEventSelectInvite(inviteIndex)
  CalendarOpenEvent(monthOffset, day, eventIndex) - Selects an existing event
  CalendarEventSetStatus(index, status) - Sets the invitation status of a player to the current event
  
  CalendarContextSelectEvent(monthOffset, day, eventButton.eventIndex);
  CalendarContextEventRemove();

  -- Query
  CalendarEventGetInvite(inviteeIndex) - Returns status information for an invitee for the currently opened event 
  CalendarEventGetNumInvites() - Returns the number of invitees for the currently opened event  
  local weekday, month, day, year, hour, minute = CalendarEventGetInviteResponseTime(self.inviteIndex);
  CalendarEventGetTypes() - Returns name1, name2, ... 
  CalendarEventIsModerator() - ?
  CalendarEventCanEdit() - Returns true if the event can be edited
  CalendarCanAddEvent() - Returns true if player can add an event 
  CalendarCanSendInvite() - Returns true if player can send invites 
  CalendarGetNumDayEvents(monthOffset[, day])
  CalendarEventGetCalendarType() -> "GUILD_ANNOUNCEMENT", "GUILD_EVENT", "PLAYER"
  CalendarEventGetTextures(eventType) -> liste des textures avec nom instance (localis !)
  local month, year, numDays, firstWeekday = CalendarGetMonth([-1,0,1]);
]]

--------------- Shared variables ---------------

--GEM3_UI_Callbacks = {};


--------------- Local variables ---------------

-- Optims vars (get a local copy)
local gmatch = string.gmatch;
local tonumber = tonumber;
local strfind = string.find;
local strformat = string.format;
local strupper = string.upper;

local _GEM3_CAL_EventFrame = nil;
local _GEM3_CAL_IsBusy = false;
local _GEM3_CAL_PendingQueries = {};
local _GEM3_CAL_IsLocked = false;
local _GEM3_CAL_StartModify = false;
local _GEM3_CAL_ModifiedEventOpened = false;
local _GEM3_CAL_NewGlobalValues = {};
local _GEM3_CAL_AssociationBuilded = false;
local _GEM3_CAL_LockedFunctions = {
  ["CalendarNewEvent"] = { orig=nil, },
  ["CalendarNewGuildEvent"] = { orig=nil, },
  ["CalendarAddEvent"] = { orig=nil, },
  ["CalendarOpenEvent"] = { orig=nil, },
  ["CalendarCloseEvent"] = { orig=nil, },
  ["CalendarUpdateEvent"] = { orig=nil, },
  ["CalendarSetAbsMonth"] = { orig=nil, },
  ["CalendarSetMonth"] = { orig=nil, },
};

local _GEM3_CAL_TexturesAssociation = {
  -- WotLK
  [GEM3_INST_TAG_NAXXRAMAS] = {texture="NAXXRAMAS",diff={}},
  [GEM3_INST_TAG_EYEOFETERNITY] = {texture="MALYGOS",diff={}},
  [GEM3_INST_TAG_CHAMBEROFASPECTS] = {texture="CHAMBEROFASPECTS",diff={}},
  [GEM3_INST_TAG_TRIAL_CRUSADER] = {texture="ARGENTRAID",diff={}},
  [GEM3_INST_TAG_ULDUAR] = {texture="ULDUAR",diff={}},
  [GEM3_INST_TAG_WINTERGRASP] = {texture="VAULTOFARCHAVON",diff={}},
  [GEM3_INST_TAG_ICC] = {texture="ICECROWNCITADEL",diff={}},
  -- TBC
  [GEM3_INST_TAG_BT] = {texture="BLACKTEMPLE",diff={}},
  [GEM3_INST_TAG_GRUUL] = {texture="GRUULSLAIR",diff={}},
  [GEM3_INST_TAG_HYJAL] = {texture="HYJALPAST",diff={}},
  [GEM3_INST_TAG_KARA] = {texture="KARAZHAN",diff={}},
  [GEM3_INST_TAG_MAGTH] = {texture="HELLFIRECITADELRAID",diff={}},
  [GEM3_INST_TAG_SSC] = {texture="SERPENTSHRINECAVERN",diff={}},
  [GEM3_INST_TAG_EYE] = {texture="TEMPESTKEEP",diff={}},
  [GEM3_INST_TAG_SWP] = {texture="SUNWELL",diff={}},
  [GEM3_INST_TAG_ZA] = {texture="ZULAMAN",diff={}},
  -- Vanilla
  [GEM3_INST_TAG_AQ20] = {texture="AQRUINS",diff={}},
  [GEM3_INST_TAG_AQ40] = {texture="AQTEMPLE",diff={}},
  [GEM3_INST_TAG_BWL] = {texture="BLACKWINGLAIR",diff={}},
  [GEM3_INST_TAG_MC] = {texture="MOLTENCORE",diff={}},
  [GEM3_INST_TAG_ONY] = {texture="",diff={}}, -- onyxia has no texture
  [GEM3_INST_TAG_UBRS] = {texture="BLACKROCKSPIRE",diff={}},
  [GEM3_INST_TAG_ZG] = {texture="ZULGURUB",diff={}},
};

local _GEM3_CAL_GEM3StateToBliz = { -- GEM states translated to Calendar states (leader's "Set Invite Status")
  [GEM3_SUB_STATE_TITULAR] = CALENDAR_INVITESTATUS_CONFIRMED,
  [GEM3_SUB_STATE_SUBSTITUTE] = CALENDAR_INVITESTATUS_STANDBY,
  [GEM3_SUB_STATE_REPLACEMENT] = CALENDAR_INVITESTATUS_STANDBY,
  [GEM3_SUB_STATE_NOT_COMING] = CALENDAR_INVITESTATUS_OUT,
  [GEM3_SUB_STATE_UNSUBSCRIBING] = CALENDAR_INVITESTATUS_OUT,
};

local _GEM3_CAL_BlizToGEM3Queue = { -- Use Tentative Status to mean NotComing
  [CALENDAR_INVITESTATUS_ACCEPTED] = GEM3_SUB_FORCE_QUEUE_TITULAR,
  [CALENDAR_INVITESTATUS_DECLINED] = GEM3_SUB_FORCE_QUEUE_NOT_COMING,
  [CALENDAR_INVITESTATUS_TENTATIVE] = GEM3_SUB_FORCE_QUEUE_NOT_COMING,
  [CALENDAR_INVITESTATUS_SIGNEDUP] = GEM3_SUB_FORCE_QUEUE_TITULAR,
  [CALENDAR_INVITESTATUS_NOT_SIGNEDUP] = GEM3_SUB_FORCE_QUEUE_NOT_COMING,
};

local _GEM3_CAL_BlizToGEM3LeadForceQueue = { -- This is for players subscribing from the CALENDAR for the first time (source will be CALENDAR), Convert Tentative for Replacement
  [CALENDAR_INVITESTATUS_CONFIRMED] = GEM3_SUB_FORCE_QUEUE_TITULAR,
  [CALENDAR_INVITESTATUS_STANDBY] = GEM3_SUB_FORCE_QUEUE_SUBSTITUTE,
  [CALENDAR_INVITESTATUS_TENTATIVE] = GEM3_SUB_FORCE_QUEUE_REPLACEMENT,
  [CALENDAR_INVITESTATUS_OUT] = GEM3_SUB_FORCE_QUEUE_NOT_COMING,
};

local _GEM3_CAL_InviteStatusToString = {
  ["UNKNOWN"] = UNKNOWN,
  [CALENDAR_INVITESTATUS_CONFIRMED] = CALENDAR_STATUS_CONFIRMED,
  [CALENDAR_INVITESTATUS_ACCEPTED] = CALENDAR_STATUS_ACCEPTED,
  [CALENDAR_INVITESTATUS_DECLINED] = CALENDAR_STATUS_DECLINED,
  [CALENDAR_INVITESTATUS_OUT] = CALENDAR_STATUS_OUT,
  [CALENDAR_INVITESTATUS_STANDBY] = CALENDAR_STATUS_STANDBY,
  [CALENDAR_INVITESTATUS_INVITED] = CALENDAR_STATUS_INVITED,
  [CALENDAR_INVITESTATUS_TENTATIVE] = CALENDAR_STATUS_TENTATIVE,
  [CALENDAR_INVITESTATUS_SIGNEDUP] = CALENDAR_STATUS_SIGNEDUP,
  [CALENDAR_INVITESTATUS_NOT_SIGNEDUP] = CALENDAR_STATUS_NOT_SIGNEDUP,
};


--------------- Internal functions ---------------

local function _GEM3_CAL_BuildCommentFieldString(ev_id,ev_comment)
  return "GEM3:"..ev_id.."--Do not edit--\n"..ev_comment;
end

local function _GEM3_CAL_CommentFieldSearchString(ev_id)
  return "^GEM3:"..ev_id.."%-%-";
end

local function _GEM3_CAL_BuildCommentFieldStringConfig(revision,comment)
  return "GEM3:Config--"..revision.."--Do not edit--\n"..comment;
end

local function _GEM3_CAL_CommentFieldSearchStringConfig()
  return "^GEM3:Config%-%-";
end

local function _GEM3_CAL_BuildAssociationTable_33(...)
  local numTextures = select("#", ...) / 4;
  if(numTextures > 0)
  then
    local pos = 1;
    for textureIndex = 1, numTextures
    do
      local title,texture,expansionlevel,difficultyName = select(pos,...);
      texture = strupper(texture);
--print(textureIndex..": "..title..": "..difficultyName..": "..tostring(texture));
--RAID_DIFFICULTY_10PLAYER = "10 Player";
--RAID_DIFFICULTY_10PLAYER_HEROIC = "10 Player (Heroic)";
--RAID_DIFFICULTY_20PLAYER = "20 Player";
--RAID_DIFFICULTY_25PLAYER = "25 Player";
--RAID_DIFFICULTY_25PLAYER_HEROIC = "25 Player (Heroic)";
--RAID_DIFFICULTY_40PLAYER = "40 Player";

      local found = false;
      for _,tab in pairs(_GEM3_CAL_TexturesAssociation)
      do
        if(tab.texture == texture)
        then
          local diff = GEM3_INST_TYPE_RAID_10; -- Defaults to RAID10 (UBRS does not have a difficulty value set)
          if(difficultyName == RAID_DIFFICULTY_10PLAYER)
          then
            diff = GEM3_INST_TYPE_RAID_10;
          elseif(difficultyName == RAID_DIFFICULTY_10PLAYER_HEROIC)
          then
            diff = GEM3_INST_TYPE_RAID_10_HEROIC;
          elseif(difficultyName == RAID_DIFFICULTY_20PLAYER)
          then
            diff = GEM3_INST_TYPE_RAID_20;
          elseif(difficultyName == RAID_DIFFICULTY_25PLAYER)
          then
            diff = GEM3_INST_TYPE_RAID_25;
          elseif(difficultyName == RAID_DIFFICULTY_25PLAYER_HEROIC)
          then
            diff = GEM3_INST_TYPE_RAID_25_HEROIC;
          elseif(difficultyName == RAID_DIFFICULTY_40PLAYER)
          then
            diff = GEM3_INST_TYPE_RAID_40;
          end
          tab.diff[diff] = textureIndex;
          found = true;
        end
      end
      if(not found)
      then
        GEM3_ChatWarning("GEM3_CAL_BuildAssociationTable: Failed to find association for "..title.." (locale: "..GetLocale()..")");
      end
      pos = pos + 4;
    end
  end
end

local function _GEM3_CAL_BuildAssociationTable_32(...)
  -- Override table
  _GEM3_CAL_TexturesAssociation = {
    -- WotLK
    [GEM3_INST_TAG_NAXXRAMAS] = {texture="Naxxramas",diff={[GEM3_INST_TYPE_RAID_10]=0,[GEM3_INST_TYPE_RAID_25]=1}},
    [GEM3_INST_TAG_EYEOFETERNITY] = {texture="Malygos",diff={[GEM3_INST_TYPE_RAID_10]=0,[GEM3_INST_TYPE_RAID_25]=1}},
    [GEM3_INST_TAG_CHAMBEROFASPECTS] = {texture="ChamberOfAspects",diff={[GEM3_INST_TYPE_RAID_10]=0,[GEM3_INST_TYPE_RAID_25]=1}},
    [GEM3_INST_TAG_TRIAL_CRUSADER] = {texture="ArgentRaid",diff={[GEM3_INST_TYPE_RAID_10]=0,[GEM3_INST_TYPE_RAID_25]=1,[GEM3_INST_TYPE_RAID_10_HEROIC]=2,[GEM3_INST_TYPE_RAID_25_HEROIC]=3}},
    [GEM3_INST_TAG_ULDUAR] = {texture="Ulduar",diff={[GEM3_INST_TYPE_RAID_10]=0,[GEM3_INST_TYPE_RAID_25]=1}},
    [GEM3_INST_TAG_WINTERGRASP] = {texture="VaultOfArchavon",diff={[GEM3_INST_TYPE_RAID_10]=0,[GEM3_INST_TYPE_RAID_25]=1}},
    -- TBC
    [GEM3_INST_TAG_BT] = {texture="BlackTemple",diff={[GEM3_INST_TYPE_RAID_25]=0}},
    [GEM3_INST_TAG_GRUUL] = {texture="GruulsLair",diff={[GEM3_INST_TYPE_RAID_25]=0}},
    [GEM3_INST_TAG_HYJAL] = {texture="HyjalPast",diff={[GEM3_INST_TYPE_RAID_25]=0}},
    [GEM3_INST_TAG_KARA] = {texture="Karazhan",diff={[GEM3_INST_TYPE_RAID_10]=0}},
    [GEM3_INST_TAG_MAGTH] = {texture="HellfireCitadelRaid",diff={[GEM3_INST_TYPE_RAID_25]=0}},
    [GEM3_INST_TAG_SSC] = {texture="SerpentshrineCavern",diff={[GEM3_INST_TYPE_RAID_25]=0}},
    [GEM3_INST_TAG_EYE] = {texture="TempestKeep",diff={[GEM3_INST_TYPE_RAID_25]=0}},
    [GEM3_INST_TAG_SWP] = {texture="Sunwell",diff={[GEM3_INST_TYPE_RAID_25]=0}},
    [GEM3_INST_TAG_ZA] = {texture="ZulAman",diff={[GEM3_INST_TYPE_RAID_10]=0}},
    -- Vanilla
    [GEM3_INST_TAG_AQ20] = {texture="AQRuins",diff={[GEM3_INST_TYPE_RAID_20]=0}},
    [GEM3_INST_TAG_AQ40] = {texture="AQTemple",diff={[GEM3_INST_TYPE_RAID_40]=0}},
    [GEM3_INST_TAG_BWL] = {texture="BlackwingLair",diff={[GEM3_INST_TYPE_RAID_40]=0}},
    [GEM3_INST_TAG_MC] = {texture="MoltenCore",diff={[GEM3_INST_TYPE_RAID_40]=0}},
    [GEM3_INST_TAG_ONY] = {texture="",diff={[GEM3_INST_TYPE_RAID_40]=0}}, -- onyxia has no texture
    [GEM3_INST_TAG_UBRS] = {texture="BlackrockSpire",diff={[GEM3_INST_TYPE_RAID_10]=0}},
    [GEM3_INST_TAG_ZG] = {texture="ZulGurub",diff={[GEM3_INST_TYPE_RAID_20]=0}},
  };

  local numTextures = select("#", ...) / 3;
  if(numTextures > 0)
  then
    local pos = 1;
    for textureIndex = 1, numTextures
    do
      local title,texture,expansionlevel = select(pos,...);

      local found = false;
      for _,tab in pairs(_GEM3_CAL_TexturesAssociation)
      do
        if(tab.texture == texture)
        then
          if(tab.base == nil)
          then
            tab.base = textureIndex;
          elseif(textureIndex < tab.base)
          then
            GEM3_ChatWarning("GEM3_CAL_BuildAssociationTable: textureIndex is smaller than stored base value for "..title.." (locale: "..GetLocale()..")");
          end
          found = true;
        end
      end
      if(not found)
      then
        GEM3_ChatWarning("GEM3_CAL_BuildAssociationTable: Failed to find association for "..title.." (locale: "..GetLocale()..")");
      end
      pos = pos + 3;
    end
  end
end

local function _GEM3_CAL_BuildAssociationTable(...)
  local num = select("#", ...);
  if(num == 93)
  then
    _GEM3_CAL_BuildAssociationTable_32(...);
  else
    _GEM3_CAL_BuildAssociationTable_33(...);
  end
end

local _GEM3_CAL_CalendarFrame_ShowEventFrame_Orig;
local function _GEM3_CAL_CalendarFrame_ShowEventFrame(frame)
  if(frame == nil or CalendarFrame:IsShown())
  then
    _GEM3_CAL_CalendarFrame_ShowEventFrame_Orig(frame);
  end
end

local function _GEM3_CAL_CheckLoadCalendar()
  if(not IsAddOnLoaded("Blizzard_Calendar")) -- Load the addon now if not loaded, since we will soon use it!
  then
    GEM3_ChatDebug(GEM3_DEBUG_CALENDAR,"GEM3_CAL_CheckLoadCalendar: Loading 'Blizzard_Calendar' addon");
    LoadAddOn("Blizzard_Calendar");
    OpenCalendar(); -- And open calendar (not the frame, but the core calendar)
    CloseCalendarMenus();
  end
  if(_GEM3_CAL_AssociationBuilded == false)
  then
    _GEM3_CAL_BuildAssociationTable(CalendarEventGetTextures(1));
    _GEM3_CAL_AssociationBuilded = true;
    -- Setup hook
    _GEM3_CAL_CalendarFrame_ShowEventFrame_Orig = CalendarFrame_ShowEventFrame;
    CalendarFrame_ShowEventFrame = _GEM3_CAL_CalendarFrame_ShowEventFrame;
  end
end

local function _GEM3_CAL_SetEventData(ev_id,ev_date,ev_place,ev_comment,max_count,ev_type,ev_inst_tag,ev_inst_type)
  local tim = date("*t",ev_date);
  
  CalendarEventSetTitle(ev_place);
  CalendarEventSetDate(tim.month,tim.day,tim.year);
  CalendarEventSetTime(tim.hour,tim.min);
  if(ev_id)
  then
    CalendarEventSetDescription(_GEM3_CAL_BuildCommentFieldString(ev_id,ev_comment));
    CalendarEventSetSize(max_count);
  else
    CalendarEventSetDescription(_GEM3_CAL_BuildCommentFieldStringConfig(1,ev_comment));
  end
  
  if(ev_type >= GEM3_EVT_TYPE_RAID and ev_type <= GEM3_EVT_TYPE_RAID_WOTLKH)
  then
    CalendarEventSetType(CALENDAR_EVENTTYPE_RAID);
    local tab = _GEM3_CAL_TexturesAssociation[ev_inst_tag];
    if(tab)
    then
      local id = tab.diff[ev_inst_type];
      if(id)
      then
        if(tab.base) -- WoW 3.2
        then
          CalendarEventSetTextureID(tab.base+id);
        else -- WoW 3.3
          CalendarEventSetTextureID(id);
        end
      end
    end
  --[[elseif(ev_type >= GEM3_EVT_TYPE_INSTANCE5 and ev_type <= GEM3_EVT_TYPE_INSTANCE5_WOTLKH)
  then
    CalendarEventSetType(CALENDAR_EVENTTYPE_DUNGEON);
    local tab = _GEM3_CAL_TexturesAssociation[ev_inst_tag];
    if(tab and tab.base)
    then
      local id = tab.diff[ev_inst_type];
      if(id)
      then
        CalendarEventSetTextureID(tab.base+id);
      end
    end]]
  elseif(ev_type == GEM3_EVT_TYPE_PVP)
  then
    CalendarEventSetType(CALENDAR_EVENTTYPE_PVP);
  elseif(ev_type >= GEM3_EVT_TYPE_GUILD_EVENT)
  then
    CalendarEventSetType(CALENDAR_EVENTTYPE_MEETING);
  else
    CalendarEventSetType(CALENDAR_EVENTTYPE_OTHER);
  end
  
  return time({year=tim.year,month=tim.month,day=tim.day,hour=tim.hour,min=tim.min});
end

local function _GEM3_CAL_ComputeMonthOffsetAndDay(ev_date)
  local dst = date("*t",ev_date);
  local src = date("*t",time());

  local months = (dst.year - src.year) * 12;
  if(dst.month > src.month)
  then
    months = months + (dst.month - src.month);
  else
    months = months - (src.month - dst.month);
  end
  return months,dst.year,dst.month,dst.day;
end

local function _GEM3_CAL_ResetSearchEvent()
  _GEM3_CAL_EventFrame.gem3_infos = GA_ReleaseTable(_GEM3_CAL_EventFrame.gem3_infos);
  CalendarCreateEventFrame:Hide();
  CalendarViewHolidayFrame:Hide();
  CalendarViewRaidFrame:Hide();
end

local function _GEM3_CAL_GetInviteeResponseTime(index)
  local _,month,day,year,hour,minute = CalendarEventGetInviteResponseTime(index);
  local tim = { year=year,month=month,day=day,hour=hour,min=minute };
  return time(tim);
end

local function _GEM3_CAL_CalendarLockedFunction()
  GEM3_ChatDebug(GEM3_DEBUG_CALENDAR,"Someone is trying to interract with the calendar while it is locked... returning");
end

local function _GEM3_CAL_QueryNextEvent()
  local infos = _GEM3_CAL_EventFrame.gem3_infos;
  if(infos == nil) -- No query in progress, check scheduled
  then
    if(#_GEM3_CAL_PendingQueries ~= 0)
    then
      infos = _GEM3_CAL_PendingQueries[1];
      tremove(_GEM3_CAL_PendingQueries,1);
      _GEM3_CAL_EventFrame.gem3_infos = infos;
    end
  end
  if(infos) -- Found a query to run
  then
    if(GEM3_CAL_IsCalendarLocked())
    then
      if(infos.bypasslock)
      then
        if(infos.findconfig)
        then
          GEM3_ChatDebug(GEM3_DEBUG_CALENDAR,"QueryNextEvent: Calendar is now locked, but bypassing lock to find config");
        else
          GEM3_ChatDebug(GEM3_DEBUG_CALENDAR,"QueryNextEvent: Calendar is now locked, but bypassing lock for event of ID "..infos.ev_id);
        end
        _GEM3_CAL_LockedFunctions["CalendarSetAbsMonth"].orig(infos.month,infos.year);
        _GEM3_CAL_LockedFunctions["CalendarOpenEvent"].orig(0,infos.day,infos.index);
      else
        GEM3_ChatDebug(GEM3_DEBUG_CALENDAR,"QueryNextEvent: Calendar is now locked, rescheduled request for event of ID "..infos.ev_id);
      end
      return;
    end
    if(infos.findconfig)
    then
      GEM3_ChatDebug(GEM3_DEBUG_CALENDAR,"QueryNextEvent: Running next scheduled request to find config ("..infos.index.."/"..infos.count..")");
    else
      GEM3_ChatDebug(GEM3_DEBUG_CALENDAR,"QueryNextEvent: Running next scheduled request for event of ID "..infos.ev_id.." ("..infos.index.."/"..infos.count..")");
    end
    CalendarSetAbsMonth(infos.month,infos.year);
    CalendarOpenEvent(0,infos.day,infos.index); -- Query next
  end
end

local function _GEM3_CAL_CheckForNextQuery()
  if(_GEM3_CAL_IsBusy == false)
  then
    local infos = _GEM3_CAL_EventFrame.gem3_infos;
    
    if(infos and infos.found) -- Check for found CB
    then
      if(infos.findconfig)
      then
        GEM3_ChatDebug(GEM3_DEBUG_CALENDAR,"Calendar OnEvent: CALENDAR_ACTION_PENDING: Opened config");
      else
        GEM3_ChatDebug(GEM3_DEBUG_CALENDAR,"Calendar OnEvent: CALENDAR_ACTION_PENDING: Opened event of ID "..infos.ev_id);
      end
      if(infos.found_cb ~= nil)
      then
        infos.found_cb(infos.ev_id,infos.year,infos.month,infos.day,infos.index,infos.userdata);
      end
      _GEM3_CAL_ResetSearchEvent(); -- Reset event data, so we can possibly start another request
    end
    _GEM3_CAL_QueryNextEvent();
  end
end

local function _GEM3_CAL_OpenEventCallback(self,event,...)
  if(event == "CALENDAR_OPEN_EVENT")
  then
    if(_GEM3_CAL_EventFrame.gem3_infos == nil) -- Not waiting for event infos
    then
      return;
    end

    local infos = _GEM3_CAL_EventFrame.gem3_infos;
    local title, description, _, _, _, _, _, _, month, day, year, hour, minute = CalendarGetEventInfo();
    local tim = { year=year,month=month,day=day,hour=hour,min=minute };
    if(infos.findconfig)
    then
      if(time(tim) == infos.ev_date and strfind(description,_GEM3_CAL_CommentFieldSearchStringConfig()) ~= nil)
      then
        infos.found = true;
        _GEM3_CAL_CheckForNextQuery(); -- Check for next query here too (like in ActionPending going to not busy), because the calendar is not always flagged as "busy" when opening events (holiday events don't lock cal)
        return; -- Stop processing
      end
    elseif(time(tim) == infos.ev_date and strfind(description,_GEM3_CAL_CommentFieldSearchString(infos.ev_id)) ~= nil)
    then
      infos.found = true;
      _GEM3_CAL_CheckForNextQuery(); -- Check for next query here too (like in ActionPending going to not busy), because the calendar is not always flagged as "busy" when opening events (holiday events don't lock cal)
      return; -- Stop processing
    end

    -- Check for new events (created one while a search is running)
    CalendarSetAbsMonth(infos.month,infos.year);
    local count = CalendarGetNumDayEvents(0,infos.day);
    if(count ~= infos.count)
    then
      GEM3_ChatDebug(GEM3_DEBUG_CALENDAR,"Calendar OnEvent: CALENDAR_OPEN_EVENT: Count of events for the day has changed from "..infos.count.." to "..count.."!! Restarting search...");
      infos.index = 1; -- Restart the search
      infos.count = count; -- Update the count
      return;
    end

    -- Next event
    if(infos.index >= infos.count)
    then
      if(infos.not_found_cb ~= nil)
      then
        infos.not_found_cb(infos.ev_id,infos.userdata);
      end
      _GEM3_CAL_ResetSearchEvent();
      _GEM3_CAL_CheckForNextQuery(); -- Check for next query here too (like in ActionPending going to not busy), because the calendar is not always flagged as "busy" when opening events (holiday events don't lock cal)
      return;
    end
    
    infos.index = infos.index + 1; -- Ready for next query
    _GEM3_CAL_CheckForNextQuery(); -- Check for next query here too (like in ActionPending going to not busy), because the calendar is not always flagged as "busy" when opening events (holiday events don't lock cal)

  elseif(event == "CALENDAR_ACTION_PENDING")
  then
    local busyFlag = select(1,...);

    _GEM3_CAL_IsBusy = busyFlag;
    _GEM3_CAL_CheckForNextQuery();
  
  elseif(event == "CALENDAR_UPDATE_INVITE_LIST")
  then
    -- Invitee list has been updated for an event (don't know which one...), we must check all events I'm the leader of
    GEM3_ChatDebug(GEM3_DEBUG_CALENDAR,"Calendar OnEvent: CALENDAR_UPDATE_INVITE_LIST: Invitee list has been updated for an event, processing events I'm the leader of");
    for ev_id,event in pairs(GEM3_QA_Events.events)
    do
      if(GEM3_IsMyReroll(event.leader))
      then
        local handle,isguild = GEM3_CAL_MustHandle(event.channel);
        if(handle) -- Using blizzard calendar?
        then
          GEM3_CAL_SynchronizeBlizzCal(ev_id);
        end
      end
    end

  elseif(event == "CALENDAR_NEW_EVENT")
  then
    GEM3_ChatDebug(GEM3_DEBUG_CALENDAR,"Calendar OnEvent: CALENDAR_NEW_EVENT: An event has been created!");
    for ev_id,event in pairs(GEM3_QA_Events.events)
    do
      if(GEM3_IsMyReroll(event.leader))
      then
        local handle,isguild = GEM3_CAL_MustHandle(event.channel);
        if(handle) -- Using blizzard calendar?
        then
          GEM3_CAL_SynchronizeBlizzCal(ev_id);
        end
      end
    end
  
  elseif(event == "CALENDAR_UPDATE_EVENT")
  then
    GEM3_ChatDebug(GEM3_DEBUG_CALENDAR,"Calendar OnEvent: CALENDAR_NEW_EVENT: An event has been updated!");
    -- Check global config update
    GEM3_LoadGlobalConfig();
  end
end

local function _GEM3_CAL_StartEventModify_found_cb(ev_id,year,month,day,index)
  GEM3_ChatDebug(GEM3_DEBUG_CALENDAR,"StartEventModify CB: Time to lock calendar: "..tostring(_GEM3_CAL_StartModify));
  if(_GEM3_CAL_StartModify) -- Still in modify event
  then
    _GEM3_CAL_ModifiedEventOpened = true;
    GEM3_CAL_LockCalendar();
  end
end

local function _GEM3_CAL_StartEventModify_not_found_cb(ev_id)
  GEM3_ChatDebug(GEM3_DEBUG_CALENDAR,"StartEventModify CB: Event of ID "..tostring(ev_id).." NOT FOUND!");
  if(_GEM3_CAL_StartModify) -- Still in modify event
  then
    _GEM3_CAL_StartModify = false; -- Force a cancel StartModify so that a validate won't bloc
  end
  if(GEM3_CAL_IsCalendarLocked())
  then
    GEM3_CAL_UnlockCalendar();
  end
end

local function _GEM3_CAL_LoadGlobalConfig_Found(ev_id,year,month,day,index,userdata)
  local title,description,creator = CalendarGetEventInfo();
  local previous_revision = GEM3_GlobalConfig["ConfigRevision"];
  local new_revision = nil;

  for i,line in ipairs({strsplit("\n",description)})
  do
    if(i == 1) -- Config marker, check revision
    then
      local _,_,revision = strfind(line,"^GEM3:Config%-%-(%d+)%-%-");
      if(revision == nil) -- Corrupted or not GEM3 Global Config!
      then
        return;
      end
      if(previous_revision and previous_revision == revision) -- Config already loaded, or same revision
      then
        return;
      end
      new_revision = revision;
      GEM3_GlobalConfig = {}; -- Reset config
      GEM3_GlobalConfig["ConfigRevision"] = new_revision;
    else
      if(new_revision == nil) -- Corrupted or not GEM3 Global Config!
      then
        return;
      end
      local _,_,name,value = strfind(line,"([^:]+):(.+)");
      if(name)
      then
        GEM3_GlobalConfig[name] = value;
      end
    end
  end
  -- GUI Callback
  GEM3_TriggerCallback("OnGlobalConfigLoaded",previous_revision == nil); -- Boolean 'firstLoad' indicating a first session load or a reload
  if(month == 12 and userdata == true and GEM3_CAL_CanCreateGuildEvent()) -- This is december, and the event is still for this year, and I have the rights to modify the event
  then
    GEM3_TriggerCallback("OnGlobalConfigExpiring");
  end
end

local function _GEM3_CAL_LoadGlobalConfig_NotFound(ev_id,userdata)
  if(userdata)
  then
    local tim = date("*t");
    GEM3_CAL_FindConfig(time({year=tim.year+1,month=12,day=31,hour=23,min=59}),_GEM3_CAL_LoadGlobalConfig_Found,_GEM3_CAL_LoadGlobalConfig_NotFound,false);
    return;
  end
  GEM3_TriggerCallback("OnGlobalConfigNotFound");
end

local function _GEM3_CAL_StartConfigModify_found_cb(ev_id,year,month,day,index)
  GEM3_ChatDebug(GEM3_DEBUG_CALENDAR,"StartConfigModify CB: Time to lock calendar: "..tostring(_GEM3_CAL_StartModify));
  if(_GEM3_CAL_StartModify) -- Still in modify event
  then
    _GEM3_CAL_ModifiedEventOpened = true;
    GEM3_CAL_LockCalendar();
  end
end

local function _GEM3_CAL_StartConfigModify_not_found_cb(ev_id)
  GEM3_ChatDebug(GEM3_DEBUG_CALENDAR,"StartConfigModify CB: Config not found!");
  if(_GEM3_CAL_StartModify) -- Still in modify event
  then
    if(userdata)
    then
      local tim = date("*t");
      GEM3_ChatDebug(GEM3_DEBUG_CALENDAR,"StartConfigModify CB: This is december, checking next year");
      GEM3_CAL_FindConfig(time({year=tim.year+1,month=12,day=31,hour=23,min=59}),_GEM3_CAL_StartConfigModify_found_cb,_GEM3_CAL_StartConfigModify_not_found_cb,false,true);
      return;
    end
    _GEM3_CAL_StartModify = false; -- Force a cancel StartModify so that a validate won't bloc
  end
  if(GEM3_CAL_IsCalendarLocked())
  then
    GEM3_CAL_UnlockCalendar();
  end
end

local function _GEM3_CAL_SynchronizeBlizzCal_found_cb(ev_id,year,month,day,index)
  local event = GEM3_QA_Events.events[ev_id];
  if(event)
  then
    local list_of_invitees = GA_GetTable();
    -- Check #1 - Process Calendar's list and compare with GEM's list
    local invitees = CalendarEventGetNumInvites();
    for k=1,invitees
    do
      local name,level,_,className,inviteStatus,modStatus = CalendarEventGetInvite(k);
      list_of_invitees[name] = true; -- In invitees list
      local pl_sub = event.players[name];-- GEM3_Player struct
      if(pl_sub) -- Exists
      then
        -- Checking difference btw GEM's state and current Calendar's state
        GEM3_ChatDebug(GEM3_DEBUG_CALENDAR,"SynchronizeBlizzCal: Player "..name.." already in GEM's event list, checking difference");
        local state = _GEM3_CAL_GEM3StateToBliz[pl_sub.current_queue];
        if(state)
        then
          -- State modified in Calendar
          if(inviteStatus == CALENDAR_INVITESTATUS_ACCEPTED or inviteStatus == CALENDAR_INVITESTATUS_DECLINED or inviteStatus == CALENDAR_INVITESTATUS_TENTATIVE or inviteStatus == CALENDAR_INVITESTATUS_SIGNEDUP or inviteStatus == CALENDAR_INVITESTATUS_NOT_SIGNEDUP)
          then
            local stamp = _GEM3_CAL_GetInviteeResponseTime(k);
            if(pl_sub.update_time < stamp or pl_sub.main_time < stamp)
            then
              GEM3_ChatDebug(GEM3_DEBUG_CALENDAR,"SynchronizeBlizzCal: Player "..name.." resubscribed: Dates: "..pl_sub.update_time.." - "..stamp);
              GEM3_SUB_SubscribeCalendarUpdate(ev_id,name,pl_sub.guild,level,className,_GEM3_CAL_BlizToGEM3Queue[inviteStatus],stamp,stamp,true);
            end
            CalendarEventSetStatus(k,state); -- Update state in calendar
          -- New state (modified in GEM), but not SignedUp state (meaning unsubscribed then resubscribed from Calendar)
          elseif(state ~= inviteStatus)
          then
            GEM3_ChatDebug(GEM3_DEBUG_CALENDAR,"SynchronizeBlizzCal: Must update state for "..name.." ("..k.."): "..pl_sub.current_queue.." -> "..state.." (".._GEM3_CAL_InviteStatusToString[state]..")");
            CalendarEventSetStatus(k,state);
          -- Same state, check level change (to update GEM)
          elseif(state == inviteStatus)
          then
            if(pl_sub.level ~= level)
            then
              local stamp = _GEM3_CAL_GetInviteeResponseTime(k);
              GEM3_ChatDebug(GEM3_DEBUG_CALENDAR,"SynchronizeBlizzCal: Level changed for "..name..": Old="..pl_sub.level.." New="..level);
              GEM3_SUB_SubscribeCalendarUpdate(ev_id,name,pl_sub.guild,level,className,pl_sub.force_queue,stamp,GEM3_STAMP_GetLocalStamp(),false); -- In case of level change, use 'now' as update_time
            end
          -- Unknown case: Should not happen!!
          else
            GEM3_RuntimeDebug("GEM3_CAL_SynchronizeBlizzCal: EventID "..ev_id..": Unhandled case for "..name..": "..inviteStatus.." (".._GEM3_CAL_InviteStatusToString[inviteStatus]..")");
          end
        else -- Should never happen!!!
          GEM3_RuntimeDebug("GEM3_CAL_SynchronizeBlizzCal: EventID "..ev_id..": Unknown state from "..name..": "..inviteStatus.." (".._GEM3_CAL_InviteStatusToString[inviteStatus]..")");
        end
      else -- Not in GEM list
        -- Accepted/Declined are states for manual invite in Calendar when the player responded to the invitation
        if(inviteStatus == CALENDAR_INVITESTATUS_ACCEPTED or inviteStatus == CALENDAR_INVITESTATUS_DECLINED)
        then
          local guild = ""; -- TEMPO??
          local stamp = _GEM3_CAL_GetInviteeResponseTime(k);
          local result,reason = GEM3_SUB_SubscribeCalendar(ev_id,name,guild,level,className,_GEM3_CAL_BlizToGEM3Queue[inviteStatus],stamp);
        -- SignedUp/NotSignedUp are states for Guild events, when a player saw the event in Calendar and subscribed to it
        -- Don't use Tentative here, since it can be used for manual invite too
        -- As of 3.3, it seems that NotSignedUp state no longer exists
        elseif(inviteStatus == CALENDAR_INVITESTATUS_SIGNEDUP or inviteStatus == CALENDAR_INVITESTATUS_NOT_SIGNEDUP)
        then
          local guild = GEM3_PlayerGuild; -- Signup/NotSignedup can only be from my guild
          local stamp = _GEM3_CAL_GetInviteeResponseTime(k);
          local result,reason = GEM3_SUB_SubscribeCalendar(ev_id,name,guild,level,className,_GEM3_CAL_BlizToGEM3Queue[inviteStatus],stamp);
        elseif(inviteStatus == CALENDAR_INVITESTATUS_CONFIRMED or inviteStatus == CALENDAR_INVITESTATUS_STANDBY or inviteStatus == CALENDAR_INVITESTATUS_OUT or inviteStatus == CALENDAR_INVITESTATUS_TENTATIVE)
        then
          local guild = ""; -- TEMPO??
          local stamp = _GEM3_CAL_GetInviteeResponseTime(k);
          if(stamp == nil) -- Probably myself, and if I'm not in the list, then it's because I don't want to attend the event
          then
            stamp = time();
            GEM3_ChatDebug(GEM3_DEBUG_CALENDAR,"SynchronizeBlizzCal: Player "..name.." is not in GEM's event list and has no ResponseTime: "..className.." - "..inviteStatus.." (".._GEM3_CAL_InviteStatusToString[inviteStatus].."). Forcing as not coming");
            local result,reason = GEM3_SUB_SubscribeCalendar(ev_id,name,guild,level,className,GEM3_SUB_FORCE_QUEUE_NOT_COMING,stamp,GEM3_SUB_FORCE_QUEUE_NOT_COMING);
            return;
          end
          GEM3_ChatDebug(GEM3_DEBUG_CALENDAR,"SynchronizeBlizzCal: Player "..name.." is not in GEM's event list: "..className.." - "..inviteStatus.." (".._GEM3_CAL_InviteStatusToString[inviteStatus].."). Adding him and forcing 'lead queue'");
          local result,reason = GEM3_SUB_SubscribeCalendar(ev_id,name,guild,level,className,_GEM3_CAL_BlizToGEM3LeadForceQueue[inviteStatus],stamp,_GEM3_CAL_BlizToGEM3LeadForceQueue[inviteStatus]);
        else
          GEM3_ChatDebug(GEM3_DEBUG_CALENDAR,"SynchronizeBlizzCal: Player "..name.." is not in GEM's event list: "..className.." - "..inviteStatus.." (".._GEM3_CAL_InviteStatusToString[inviteStatus]..")");
        end
      end
    end
    -- Check #2 - Process GEM's list and search for missing subscribers in Calendar's list
    for name,pl_sub in pairs(event.players) -- GEM3_Player struct
    do
      if(pl_sub.source == GEM3_SUB_SOURCE_CALENDAR and pl_sub.current_queue ~= GEM3_SUB_STATE_NOT_COMING) -- Source is calendar, so we should find this player in calendar
      then
        if(list_of_invitees[name] == nil) -- No longer in invitees list
        then
          GEM3_ChatDebug(GEM3_DEBUG_CALENDAR,"SynchronizeBlizzCal: Player "..name.." was added from Calendar, but is no longer in Calendar's list, remove him from GEM");
          stamp = time();
          GEM3_SUB_SubscribeCalendarUpdate(ev_id,name,pl_sub.guild,pl_sub.level,pl_sub.class,GEM3_SUB_FORCE_QUEUE_NOT_COMING,stamp,stamp,true);
        end
      end
    end
    GA_ReleaseTable(list_of_invitees);
  end
end

local function _GEM3_CAL_SynchronizeBlizzCal_not_found_cb(ev_id)
  GEM3_ChatDebug(GEM3_DEBUG_CALENDAR,"Event "..ev_id.." not found when trying to synchronize with blizzard calendar");
end

local function _GEM3_CAL_DeleteEvent_found_cb(ev_id,year,month,day,index)
  --CalendarContextSelectEvent(monthOffset,day,index);
  CalendarSetAbsMonth(month,year);
  CalendarContextEventRemove(0,day,index);
  GEM3_ChatDebug(GEM3_DEBUG_CALENDAR,"DeleteEvent: Removed event of ID "..ev_id);
end

local function _GEM3_CAL_DeleteEvent_not_found_cb(ev_id)
  GEM3_ChatDebug(GEM3_DEBUG_CALENDAR,"Event "..ev_id.." not found when trying to delete it from blizzard calendar");
end


--------------- Initialization ---------------

_GEM3_CAL_EventFrame = CreateFrame("Frame");
_GEM3_CAL_EventFrame:RegisterEvent("CALENDAR_OPEN_EVENT");
_GEM3_CAL_EventFrame:RegisterEvent("CALENDAR_ACTION_PENDING");
_GEM3_CAL_EventFrame:RegisterEvent("CALENDAR_UPDATE_INVITE_LIST");
_GEM3_CAL_EventFrame:RegisterEvent("CALENDAR_NEW_EVENT");
_GEM3_CAL_EventFrame:RegisterEvent("CALENDAR_UPDATE_EVENT");
_GEM3_CAL_EventFrame:SetScript("OnEvent",_GEM3_CAL_OpenEventCallback);


--------------- Exported functions ---------------

function GEM3_CAL_CreateNewCustomEvent(ev_id,ev_date,ev_place,ev_comment,max_count,min_lvl,max_lvl,ev_type,ev_inst_tag,ev_inst_type)
  local offset = _GEM3_CAL_ComputeMonthOffsetAndDay(ev_date);
  if(offset >= 13)
  then
    return false,GEM3_CAL_ERROR_TOO_FAR;
  end
  if(_GEM3_CAL_IsBusy) -- Calendar is busy, return error
  then
    return false,GEM3_CAL_ERROR_BUSY;
  end
  CalendarNewEvent();
  local created_time = _GEM3_CAL_SetEventData(ev_id,ev_date,ev_place,ev_comment,max_count,ev_type,ev_inst_tag,ev_inst_type);
  CalendarAddEvent(); -- Flush the event
  GEM3_ChatDebug(GEM3_DEBUG_CALENDAR,"GEM3_CAL_CreateNewCustomEvent: Creating custom event of ID "..ev_id.." at time "..created_time);
  return true,created_time;
end

function GEM3_CAL_CreateNewGuildEvent(ev_id,ev_date,ev_place,ev_comment,max_count,min_lvl,max_lvl,ev_type,ev_inst_tag,ev_inst_type)
  if(GEM3_CAL_CanCreateGuildEvent() == false)
  then
    GEM3_ChatDebug(GEM3_DEBUG_CALENDAR,"GEM3_CAL_CreateNewGuildEvent: You don't have the rights to create blizzard guild events, for event of ID "..ev_id);
    return false,GEM3_CAL_ERROR_NO_RIGHTS;
  end
  local offset = _GEM3_CAL_ComputeMonthOffsetAndDay(ev_date);
  if(offset >= 13)
  then
    return false,GEM3_CAL_ERROR_TOO_FAR;
  end
  if(_GEM3_CAL_IsBusy) -- Calendar is busy, return error
  then
    return false,GEM3_CAL_ERROR_BUSY;
  end
  CalendarNewGuildEvent(min_lvl,max_lvl);
  local created_time = _GEM3_CAL_SetEventData(ev_id,ev_date,ev_place,ev_comment,max_count,ev_type,ev_inst_tag,ev_inst_type);
  CalendarAddEvent(); -- Flush the event
  GEM3_ChatDebug(GEM3_DEBUG_CALENDAR,"GEM3_CAL_CreateNewGuildEvent: Creating guild event of ID "..ev_id.." at time "..created_time);
  return true,created_time;
end

function GEM3_CAL_CreateNewGuildAnnouncement(ev_date,ev_place,ev_comment,ev_type,ev_inst_tag,ev_inst_type)
  if(GEM3_CAL_CanCreateGuildEvent() == false)
  then
    GEM3_ChatDebug(GEM3_DEBUG_CALENDAR,"GEM3_CAL_CreateNewGuildAnnouncement: You don't have the rights to create blizzard guild events, for event of ID "..ev_id);
    return false,GEM3_CAL_ERROR_NO_RIGHTS;
  end
  local offset = _GEM3_CAL_ComputeMonthOffsetAndDay(ev_date);
  if(offset >= 13)
  then
    return false,GEM3_CAL_ERROR_TOO_FAR;
  end
  if(_GEM3_CAL_IsBusy) -- Calendar is busy, return error
  then
    return false,GEM3_CAL_ERROR_BUSY;
  end
  CalendarNewGuildAnnouncement();
  local created_time = _GEM3_CAL_SetEventData(ev_id,ev_date,ev_place,ev_comment,max_count,ev_type,ev_inst_tag,ev_inst_type);
  CalendarAddEvent(); -- Flush the event
  GEM3_ChatDebug(GEM3_DEBUG_CALENDAR,"GEM3_CAL_CreateNewGuildAnnouncement: Creating guild announcement at time "..created_time);
  return true,created_time;
end

function GEM3_CAL_CanCreateGuildEvent()
  return CanEditGuildEvent() == 1;
end

function GEM3_CAL_FindEvent(ev_id,ev_date,found_cb,not_found_cb,userdata,bypasslock)
  _GEM3_CAL_CheckLoadCalendar();

  local _,year,month,day = _GEM3_CAL_ComputeMonthOffsetAndDay(ev_date);
  local count = 0;

  if(GEM3_CAL_IsCalendarLocked())
  then
    if(bypasslock)
    then
      GEM3_ChatDebug(GEM3_DEBUG_CALENDAR,"GEM3_CAL_FindEvent: Calendar is currently locked, but bypassing lock for event of ID "..ev_id);
      _GEM3_CAL_LockedFunctions["CalendarSetAbsMonth"].orig(month,year);
      count = CalendarGetNumDayEvents(0,day);
    end
  else
    CalendarSetAbsMonth(month,year);
    count = CalendarGetNumDayEvents(0,day);
  end

  if(count == 0 and not GEM3_CAL_IsCalendarLocked()) -- To prevent returning right now if the calendar is locked, we need to schedule the query
  then
    GEM3_ChatDebug(GEM3_DEBUG_CALENDAR,"GEM3_CAL_FindEvent: No event found on calendar for event ID "..ev_id);
    return false,"No events";
  end

  -- Search for same query already scheduled
  for i,infos in ipairs(_GEM3_CAL_PendingQueries)
  do
    if(infos.ev_id == ev_id and infos.found_cb == found_cb and infos.not_found_cb == not_found_cb and infos.userdata == userdata and infos.bypasslock == bypasslock)
    then
      GEM3_ChatDebug(GEM3_DEBUG_CALENDAR,"GEM3_CAL_FindEvent: Already a pending request for event ID "..ev_id.."... not scheduling a new identical search");
      return true;
    end
  end
  
  local insert_pos = #_GEM3_CAL_PendingQueries + 1;
  if(bypasslock)
  then
    insert_pos = 1;
  end

  local infos = GA_GetTable();
  infos.count = count;
  infos.ev_id = ev_id;
  infos.ev_date = ev_date;
  infos.year = year;
  infos.month = month;
  infos.day = day;
  infos.index = 1;
  infos.found_cb = found_cb;
  infos.not_found_cb = not_found_cb;
  infos.userdata = userdata;
  infos.bypasslock = bypasslock;
  
  if(_GEM3_CAL_IsBusy) -- Calendar is busy, we must schedule query
  then
    tinsert(_GEM3_CAL_PendingQueries,insert_pos,infos);
    GEM3_ChatDebug(GEM3_DEBUG_CALENDAR,"GEM3_CAL_FindEvent: Calendar is busy, scheduling query for event of ID "..ev_id);
    return false,"Calendar is busy, scheduling query";
  end
  
  if(_GEM3_CAL_EventFrame.gem3_infos ~= nil) -- Already a pending query,must schedule
  then
    tinsert(_GEM3_CAL_PendingQueries,insert_pos,infos);
    GEM3_ChatDebug(GEM3_DEBUG_CALENDAR,"GEM3_CAL_FindEvent: Already a pending request, scheduling query for event of ID "..ev_id);
    return false,"Request already pending, scheduling query";
  end

  if(GEM3_CAL_IsCalendarLocked())
  then
    if(bypasslock)
    then
      GEM3_ChatDebug(GEM3_DEBUG_CALENDAR,"GEM3_CAL_FindEvent: Calendar is currently locked, but bypassing lock and opening event of ID "..ev_id);
      _GEM3_CAL_EventFrame.gem3_infos = infos;
      _GEM3_CAL_LockedFunctions["CalendarOpenEvent"].orig(0,infos.day,infos.index); -- SetAbsMonth already done
      return true;
    else
      tinsert(_GEM3_CAL_PendingQueries,insert_pos,infos);
      GEM3_ChatDebug(GEM3_DEBUG_CALENDAR,"GEM3_CAL_FindEvent: Calendar is currently locked, scheduling query for event of ID "..ev_id);
      return false,"Calendar currently locked, scheduling query";
    end
  end

  _GEM3_CAL_EventFrame.gem3_infos = infos;
  CalendarOpenEvent(0,infos.day,infos.index); -- SetAbsMonth already done
  return true;
end

function GEM3_CAL_FindConfig(ev_date,found_cb,not_found_cb,userdata,bypasslock)
  _GEM3_CAL_CheckLoadCalendar();

  local _,year,month,day = _GEM3_CAL_ComputeMonthOffsetAndDay(ev_date);
  CalendarSetAbsMonth(month,year);
  local count = CalendarGetNumDayEvents(0,day);

  if(count == 0)
  then
    GEM3_ChatDebug(GEM3_DEBUG_CALENDAR,"GEM3_CAL_FindConfig: No event found on calendar for year "..year);
    return false;
  end

  -- Search for same query already scheduled
  if(_GEM3_CAL_EventFrame.gem3_infos ~= nil and _GEM3_CAL_EventFrame.gem3_infos.findconfig and _GEM3_CAL_EventFrame.gem3_infos.userdata == userdata and _GEM3_CAL_EventFrame.gem3_infos.bypasslock == bypasslock)
  then
    GEM3_ChatDebug(GEM3_DEBUG_CALENDAR,"GEM3_CAL_FindConfig: Already a running request to find config... not scheduling a new identical search");
    return true;
  end
  for i,infos in ipairs(_GEM3_CAL_PendingQueries)
  do
    if(infos.findconfig and infos.userdata == userdata and infos.bypasslock == bypasslock)
    then
      GEM3_ChatDebug(GEM3_DEBUG_CALENDAR,"GEM3_CAL_FindConfig: Already a pending request to find config... not scheduling a new identical search");
      return true;
    end
  end
  
  local insert_pos = #_GEM3_CAL_PendingQueries + 1;
  if(bypasslock)
  then
    insert_pos = 1;
  end

  local infos = GA_GetTable();
  infos.findconfig = true;
  infos.count = count;
  infos.ev_date = ev_date;
  infos.year = year;
  infos.month = month;
  infos.day = day;
  infos.index = 1;
  infos.found_cb = found_cb;
  infos.not_found_cb = not_found_cb;
  infos.userdata = userdata;
  infos.bypasslock = bypasslock;
  
  if(_GEM3_CAL_IsBusy) -- Calendar is busy, we must schedule query
  then
    tinsert(_GEM3_CAL_PendingQueries,insert_pos,infos);
    GEM3_ChatDebug(GEM3_DEBUG_CALENDAR,"GEM3_CAL_FindConfig: Calendar is busy, scheduling query to find config");
    return true;
  end
  
  if(_GEM3_CAL_EventFrame.gem3_infos ~= nil) -- Already a pending query,must schedule
  then
    tinsert(_GEM3_CAL_PendingQueries,insert_pos,infos);
    GEM3_ChatDebug(GEM3_DEBUG_CALENDAR,"GEM3_CAL_FindConfig: Already a pending request, scheduling query to find config");
    return true;
  end

  if(GEM3_CAL_IsCalendarLocked())
  then
    if(bypasslock)
    then
      GEM3_ChatDebug(GEM3_DEBUG_CALENDAR,"GEM3_CAL_FindConfig: Calendar is currently locked, but bypassing lock to find config");
      _GEM3_CAL_EventFrame.gem3_infos = infos;
      _GEM3_CAL_LockedFunctions["CalendarOpenEvent"].orig(0,infos.day,infos.index); -- SetAbsMonth already done
      return true;
    else
      tinsert(_GEM3_CAL_PendingQueries,insert_pos,infos);
      GEM3_ChatDebug(GEM3_DEBUG_CALENDAR,"GEM3_CAL_FindConfig: Calendar is currently locked, scheduling query to find config");
      return true;
    end
  end
  _GEM3_CAL_EventFrame.gem3_infos = infos;
  CalendarOpenEvent(0,infos.day,infos.index); -- SetAbsMonth already done
  return true;
end

function GEM3_CAL_CanCreateEvents(channel)
  local infos = GEM3_CHAN_GetChannelInfos(channel);
  if(infos)
  then
    if(infos.guild)
    then
      if(GEM3_CAL_CanCreateGuildEvent())
      then
        return true,infos;
      else -- Cannot create guild events
        return false,infos;
      end
    else -- Not guild channel
      return false,infos;
    end
  end
  return false,nil;
end

--[[ 
  Returns:
   handle: bool -- Must be handled by blizz calendar
   isguild: bool -- Is guild channel (only set if handle is true)
]]
function GEM3_CAL_MustHandle(channel)
  local cancreate,infos = GEM3_CAL_CanCreateEvents(channel);
  if(infos) -- Channel is known
  then
    if(infos.use_bliz_cal and cancreate)
    then
      _GEM3_CAL_CheckLoadCalendar();
      return true,infos.guild == true;
    else
      GEM3_CAL_EnableCalendar(channel,false); -- Force disable, just in case
    end
  end
  return false,false;
end

--[[ 
  Returns:
   enable: bool -- New state for the channel
]]
function GEM3_CAL_EnableCalendar(channel,enable)
  local cancreate,infos = GEM3_CAL_CanCreateEvents(channel);
  if(infos) -- Channel is known
  then
    if(enable and cancreate) -- Enable and can create events
    then
      infos.use_bliz_cal = true;
      _GEM3_CAL_CheckLoadCalendar();
      return true; -- Ok, enabled
    else -- Disable or can't create events
      if(infos.use_bliz_cal) -- Was enabled
      then
        -- TODO: Destroy my calendar events
      end
      infos.use_bliz_cal = false;
      return false; -- Ok, disabled
    end
  end
  return false;
end

function GEM3_CAL_StartEventModify(ev_id)
  local event = GEM3_QA_Events.events[ev_id];

  if(event == nil)
  then
    return;
  end

  _GEM3_CAL_StartModify = true;
  _GEM3_CAL_ModifiedEventOpened = false;
  if(_GEM3_CAL_IsBusy == false and _GEM3_CAL_EventFrame.gem3_infos == nil) -- Calendar is available
  then
    GEM3_ChatDebug(GEM3_DEBUG_CALENDAR,"GEM3_CAL_StartEventModify: Calendar available, locking it right now");
    GEM3_CAL_LockCalendar();
  end
  
  GEM3_CAL_FindEvent(ev_id,event.ev_date,_GEM3_CAL_StartEventModify_found_cb,_GEM3_CAL_StartEventModify_not_found_cb,nil,true);
end

function GEM3_CAL_CancelEventModify()
  if(_GEM3_CAL_StartModify)
  then
    _GEM3_CAL_StartModify = false;
    GEM3_ChatDebug(GEM3_DEBUG_CALENDAR,"GEM3_CAL_CancelEventModify");
    GEM3_CAL_UnlockCalendar();
  end
end

function GEM3_CAL_ValidateEventModify(ev_id,ev_date,ev_place,ev_comment,max_count,min_lvl,max_lvl,ev_type,ev_inst_tag,ev_inst_type)
  if(_GEM3_CAL_StartModify)
  then
    if(_GEM3_CAL_ModifiedEventOpened == false) -- Event not opened yet
    then
      GEM3_ChatDebug(GEM3_DEBUG_CALENDAR,"GEM3_CAL_ValidateEventModify: Event not found yet, try again later");
      return false,GEM3_CAL_ERROR_BUSY;
    end
    local offset = _GEM3_CAL_ComputeMonthOffsetAndDay(ev_date);
    if(offset >= 13)
    then
      return false,GEM3_CAL_ERROR_TOO_FAR;
    end
    _GEM3_CAL_StartModify = false;
    _GEM3_CAL_SetEventData(ev_id,ev_date,ev_place,ev_comment,max_count,ev_type,ev_inst_tag,ev_inst_type);
    GEM3_ChatDebug(GEM3_DEBUG_CALENDAR,"GEM3_CAL_ValidateEventModify: Flushing calendar event");
    if(GEM3_CAL_IsCalendarLocked())
    then
      _GEM3_CAL_LockedFunctions["CalendarUpdateEvent"].orig();
    else
      CalendarUpdateEvent();
    end
    GEM3_CAL_UnlockCalendar();
  end
  return true;
end

function GEM3_CAL_SetNewGlobalValue(name,value)
  if(_GEM3_CAL_StartModify == false)
  then
    return false;
  end
  _GEM3_CAL_NewGlobalValues[name] = value;
  return true;
end

function GEM3_CAL_StartConfigModify()
  _GEM3_CAL_StartModify = true;
  _GEM3_CAL_ModifiedEventOpened = false;
  _GEM3_CAL_NewGlobalValues = {}; -- Drop previously modified values
  if(_GEM3_CAL_IsBusy == false and _GEM3_CAL_EventFrame.gem3_infos == nil) -- Calendar is available
  then
    GEM3_ChatDebug(GEM3_DEBUG_CALENDAR,"GEM3_CAL_StartConfigModify: Calendar available, locking it right now");
    GEM3_CAL_LockCalendar();
  end
  
  local tim = date("*t");
  GEM3_CAL_FindConfig(time({year=tim.year,month=12,day=31,hour=23,min=59}),_GEM3_CAL_StartConfigModify_found_cb,_GEM3_CAL_StartConfigModify_not_found_cb,tim.month == 12,true);
end

function GEM3_CAL_CancelConfigModify()
  if(_GEM3_CAL_StartModify)
  then
    _GEM3_CAL_StartModify = false;
    _GEM3_CAL_NewGlobalValues = {}; -- Drop previously modified values
    GEM3_ChatDebug(GEM3_DEBUG_CALENDAR,"GEM3_CAL_CancelConfigModify");
    GEM3_CAL_UnlockCalendar();
  end
end

function GEM3_CAL_ValidateConfigModify(revision)
  if(_GEM3_CAL_StartModify)
  then
    if(_GEM3_CAL_ModifiedEventOpened == false) -- Event not opened yet
    then
      GEM3_ChatDebug(GEM3_DEBUG_CALENDAR,"GEM3_CAL_ValidateConfigModify: Config not found yet, try again later");
      return false,GEM3_CAL_ERROR_BUSY;
    end
    _GEM3_CAL_StartModify = false;
    
    local comment = "";
    for name,value in pairs(GEM3_GlobalConfig) -- Add non modified values
    do
      if(name ~= "ConfigRevision" and _GEM3_CAL_NewGlobalValues[name] == nil)
      then
        comment = comment .. name .. ":" .. value .. "\n";
      end
    end
    for name,value in pairs(_GEM3_CAL_NewGlobalValues) -- Add new values
    do
      comment = comment .. name .. ":" .. value .. "\n";
    end
    if(strlen(comment) > 200)
    then
      GEM3_ChatWarning("Global config exceed 200 chars");
    end
    
    CalendarEventSetDescription(_GEM3_CAL_BuildCommentFieldStringConfig(revision,comment));
    GEM3_ChatDebug(GEM3_DEBUG_CALENDAR,"GEM3_CAL_ValidateConfigModify: Flushing calendar config");
    if(GEM3_CAL_IsCalendarLocked())
    then
      _GEM3_CAL_LockedFunctions["CalendarUpdateEvent"].orig();
    else
      CalendarUpdateEvent();
    end
    GEM3_CAL_UnlockCalendar();
  end
  return true;
end

function GEM3_CAL_ValidateConfigShift()
  if(_GEM3_CAL_StartModify)
  then
    if(_GEM3_CAL_ModifiedEventOpened == false) -- Event not opened yet
    then
      GEM3_ChatDebug(GEM3_DEBUG_CALENDAR,"GEM3_CAL_ValidateConfigShift: Config not found yet, try again later");
      return false,GEM3_CAL_ERROR_BUSY;
    end
    _GEM3_CAL_StartModify = false;
    local tim = date("*t");
    CalendarEventSetDate(12,31,tim.year+1);
    GEM3_ChatDebug(GEM3_DEBUG_CALENDAR,"GEM3_CAL_ValidateConfigShift: Shifting calendar config to next year");
    if(GEM3_CAL_IsCalendarLocked())
    then
      _GEM3_CAL_LockedFunctions["CalendarUpdateEvent"].orig();
    else
      CalendarUpdateEvent();
    end
    GEM3_CAL_UnlockCalendar();
  end
  return true;
end

function GEM3_CAL_IsCalendarLocked()
  return _GEM3_CAL_IsLocked;
end

function GEM3_CAL_LockCalendar()
  _GEM3_CAL_CheckLoadCalendar();
  if(GEM3_CAL_IsCalendarLocked())
  then
    GEM3_ChatDebug(GEM3_DEBUG_CALENDAR,"GEM3_CAL_LockCalendar: Calendar already locked");
    return false;
  end

  GEM3_ChatDebug(GEM3_DEBUG_CALENDAR,"## Locking calendar");
  for name,tab in pairs(_GEM3_CAL_LockedFunctions)
  do
    tab.orig = _G[name];
    _G[name] = _GEM3_CAL_CalendarLockedFunction;
  end

  _GEM3_CAL_IsLocked = true;
  return true;
end

function GEM3_CAL_UnlockCalendar()
  if(not GEM3_CAL_IsCalendarLocked())
  then
    GEM3_ChatDebug(GEM3_DEBUG_CALENDAR,"GEM3_CAL_UnlockCalendar: Calendar not locked");
    return false;
  end

  GEM3_ChatDebug(GEM3_DEBUG_CALENDAR,"## Unlocking calendar");
  for name,tab in pairs(_GEM3_CAL_LockedFunctions)
  do
    _G[name] = tab.orig;
    tab.orig = nil;
  end

  _GEM3_CAL_IsLocked = false;
  _GEM3_CAL_QueryNextEvent(); -- Check for next query to run
  return true;
end

function GEM3_CAL_SynchronizeBlizzCal(ev_id)
  local event = GEM3_QA_Events.events[ev_id];

  GEM3_ChatDebug(GEM3_DEBUG_CALENDAR,"GEM3_CAL_SynchronizeBlizzCal: Synchronizing event of ID "..ev_id.." (Event="..tostring(event)..")");
  if(event)
  then
    GEM3_CAL_FindEvent(event.id,event.ev_date,_GEM3_CAL_SynchronizeBlizzCal_found_cb,_GEM3_CAL_SynchronizeBlizzCal_not_found_cb);
  end
end

function GEM3_CAL_DeleteEvent(ev_id)
  local event = GEM3_QA_Events.events[ev_id];

  GEM3_ChatDebug(GEM3_DEBUG_CALENDAR,"GEM3_CAL_DeleteEvent: Deleting event of ID "..ev_id.." (Event="..tostring(event)..")");
  if(event)
  then
    GEM3_CAL_FindEvent(event.id,event.ev_date,_GEM3_CAL_DeleteEvent_found_cb,_GEM3_CAL_DeleteEvent_not_found_cb,event.ev_date);
  end
end

function GEM3_CAL_LoadGlobalConfig()
  local tim = date("*t");
  GEM3_CAL_FindConfig(time({year=tim.year,month=12,day=31,hour=23,min=59}),_GEM3_CAL_LoadGlobalConfig_Found,_GEM3_CAL_LoadGlobalConfig_NotFound,tim.month == 12);
end


----------------- DEBUG
function gem3_found_cb(ev_id)
  kud("FOUND event "..ev_id);

  kud("Infos about event "..ev_id);
  local title, description, creator, eventType, repeatOption, maxSize, textureIndex, weekday, month, day, year, hour, minute,
    lockoutWeekday, lockoutMonth, lockoutDay, lockoutYear, lockoutHour, lockoutMinute,
    locked, autoApprove, pendingInvite, inviteStatus, inviteType, calendarType = CalendarGetEventInfo();
  kud(" - Title: "..tostring(title));
  kud(" - Description: "..tostring(description));
  kud(" - Creator: "..tostring(creator));
  kud(" - EventType: "..tostring(eventType));
  kud(" - MaxSize: "..tostring(maxSize));
  kud(" - TextIdx: "..tostring(textureIndex));
  local tim = { year=year,month=month,day=day,hour=hour,min=minute};
  kud(" - EvDate: "..tostring(time(tim)).." -> "..tostring(date("%c",time(tim))));
  if(lockoutYear ~= 2000)
  then
    local tim = { year=lockoutYear,month=lockoutMonth,day=lockoutDay,hour=lockoutHour,min=lockoutMinute};
    kud(" - LockoutDate: "..tostring(time(tim)).." -> "..tostring(date("%c",time(tim))));
  end
  kud(" - Locked: "..tostring(locked));
  kud(" - CalType: "..tostring(calendarType));
  --kud(" - Type: "..tostring(CalendarEventGetCalendarType())); <- comme calendarType
  kud(" - CanEdit: "..tostring(CalendarEventCanEdit()));
  kud(" - CanSendInvite: "..tostring(CalendarCanSendInvite()));
  kud(" - IsModerator: "..tostring(CalendarEventIsModerator()));
  kud(" - Num invitees: "..tostring(CalendarEventGetNumInvites()));
  local invitees = CalendarEventGetNumInvites();
  for k=1,invitees
  do
    local name, level, className, classFilename, inviteStatus = CalendarEventGetInvite(k);
    kud("  - Invitee "..name.." ["..level.."] "..className.." : ".._GEM3_CAL_InviteStatusToString[inviteStatus]);
  end
end

function gem3_not_found_cb(ev_id,userdata)
  kud("Failed to find event "..ev_id);
end

function gem3_conf_found(ev_id,year,month,day,index,userdata)
  kud("FOUND config");
  local title, description, creator, eventType, repeatOption, maxSize, textureIndex, weekday, month, day, year, hour, minute,
    lockoutWeekday, lockoutMonth, lockoutDay, lockoutYear, lockoutHour, lockoutMinute,
    locked, autoApprove, pendingInvite, inviteStatus, inviteType, calendarType = CalendarGetEventInfo();
  kud(" - Title: "..tostring(title));
  kud(" - Description: "..tostring(description));
  kud(" - Creator: "..tostring(creator));
end
 
function gem3_conf_not_found()
  kud("Failed to find config");
end

-- /run GEM3_CAL_FindConfig(time({year=2009,month=12,day=31,hour=23,min=59}),gem3_conf_found,gem3_conf_not_found)

-- /run GEM3_CAL_CreateNewGuildEvent("Kikipik1",time()+1000,"Test Event - Imported from GEM3","Test",25,70,80,GEM3_EVT_TYPE_RAID_WOTLK,GEM3_INST_TAG_TRIAL_CRUSADER,GEM3_INST_TYPE_RAID_25_HEROIC)
-- /run GEM3_CAL_FindEvent("Kirash2.428",time(),gem3_found_cb,gem3_not_found_cb);
-- /run GEM3_CAL_FindEvent("Kikipik1",1251837720,gem3_found_cb,gem3_not_found_cb,1)
-- /run GEM3_CAL_SynchronizeBlizzCal("Kirash12.978");
-- /run GEM3_CAL_FindEvent("Kikipik1",time());GEM3_CAL_FindEvent("Kikipik2",time());GEM3_CAL_LockCalendar();
-- /run GEM3_CAL_StartEventModify("Kirash9.689");print(GEM3_CAL_ValidateEventModify());
-- /run print(GEM3_CAL_EnableCalendar(GEM3_GuildChannelName,true))
-- /run print(GEM3_CAL_MustHandle(GEM3_GuildChannelName))
-- /run GEM3_CAL_SynchronizeBlizzCal("Kirash2.428")
