--[[
  Guild Event Manager (GEM) - Version 3
  by Kiki from European "Conseil des Ombres" (Horde)
  ----------------------------------------
  Constantes
]]


-- GEM3 version
GEM3_VERSION = GetAddOnMetadata("GEM3_Core", "Version");
GEM3_COM_PROTOCOL_REVISION = 3.50;
GEM3_MIN_CONFIG_VERSION_RESET = 1.29;

-- Remove player if not logged since this time
GEM3_PLAY_LASTLOG_MAX = 60*60*24*30; -- 30 days
-- Delay btw 2 players info update
GEM3_PLAY_UPDATE_INFOS_DELAY = 45; -- 45 sec
-- Delay (min/max range) before event/command broadcast for non channel leader
GEM3_QUE_SCHEDULE_DELAY_MIN = 30;
GEM3_QUE_SCHEDULE_DELAY_MAX = 90;
-- Delay between logout/login of the same player, so we don't broadcast again
GEM3_QUE_MIN_DELAY_LOGOUT_LOGIN = 60;
-- Max timestamp offset btw TimeRef and other player
GEM3_STAMP_MAX_OFFSET = 60*15; -- 15 min
GEM3_COMM_PREFIX = "GEM3";

-- Core suspend reasons
GEM3_SUSPEND_MANUAL = 0;
GEM3_SUSPEND_IN_INSTANCE = 1;

-- Subscription source
GEM3_SUB_SOURCE_SELF = "s";
GEM3_SUB_SOURCE_EXTERNAL = "e"; -- Name of the player who externally added this player, is concatenated
GEM3_SUB_SOURCE_WEB = "w"; -- Name of the player who subscribed this player, is concatenated
GEM3_SUB_SOURCE_TEMPLATE = "t";
GEM3_SUB_SOURCE_CALENDAR = "c"; -- From blizzard calendar

-- Subscription queue wanted
GEM3_SUB_FORCE_QUEUE_NONE = 0;
GEM3_SUB_FORCE_QUEUE_TITULAR = 1;
GEM3_SUB_FORCE_QUEUE_SUBSTITUTE = 2;
GEM3_SUB_FORCE_QUEUE_REPLACEMENT = 3;
GEM3_SUB_FORCE_QUEUE_NOT_COMING = 4;

-- Subscription State - used by GEM3_Player struct (server side) and GEM3_Subscription struct (client side)
GEM3_SUB_STATE_NONE = 0;
GEM3_SUB_STATE_TITULAR = 1;
GEM3_SUB_STATE_SUBSTITUTE = 2;
GEM3_SUB_STATE_REPLACEMENT = 3;
GEM3_SUB_STATE_NOT_COMING = 4;
GEM3_SUB_STATE_NOT_ACKED = 5;
GEM3_SUB_STATE_DISCARD = 6; -- Internal
GEM3_SUB_STATE_UNSUBSCRIBING = 7;

-- Subscription error codes
GEM3_SUB_ERROR_LEVEL_RANGE = 1;
GEM3_SUB_ERROR_BANNED = 2;
GEM3_SUB_ERROR_KICKED = 3;
GEM3_SUB_ERROR_UNKNOWN_EVENT = 4;
GEM3_SUB_ERROR_DEADLINE = 5;

-- Calendar error codes
GEM3_CAL_ERROR_NO_RIGHTS = 1;
GEM3_CAL_ERROR_BUSY = 2;
GEM3_CAL_ERROR_TOO_FAR = 3;

-- Class Role
GEM3_ROLE_UNKNOWN = 0;
GEM3_ROLE_TANK = 1;
GEM3_ROLE_CDPS = 2;
GEM3_ROLE_RDPS = 4;
GEM3_ROLE_DPS = 6;
GEM3_ROLE_HEAL = 8;

-- Event types
GEM3_EVT_TYPE_NONE = 0;
GEM3_EVT_TYPE_TEST = 1;
GEM3_EVT_TYPE_RESET = 2;
GEM3_EVT_TYPE_TEMPLATE = 3;
GEM3_EVT_TYPE_ARCHIVE = 4;
GEM3_EVT_TYPE_LFG = 10;
GEM3_EVT_TYPE_POLL = 11;
GEM3_EVT_TYPE_PLAYER_AVALABILITY = 12;
GEM3_EVT_TYPE_RAID_AVAILABILITY = 13;
GEM3_EVT_TYPE_INSTANCE5 = 20;
GEM3_EVT_TYPE_INSTANCE5_BC = 21;
GEM3_EVT_TYPE_INSTANCE5_WOTLK = 22;
GEM3_EVT_TYPE_INSTANCE5_BCH = 23; -- No meaning post WoW 3.2
GEM3_EVT_TYPE_INSTANCE5_WOTLKH = 24; -- No meaning post WoW 3.2
GEM3_EVT_TYPE_RAID = 30;
GEM3_EVT_TYPE_RAID_BC = 31;
GEM3_EVT_TYPE_RAID_WOTLK = 32;
GEM3_EVT_TYPE_RAID_WOTLKH = 33; -- No meaning post WoW 3.2
GEM3_EVT_TYPE_PVP = 40;
GEM3_EVT_TYPE_GUILD_EVENT = 60;
GEM3_EVT_TYPE_CUSTOM_START = 100; -- Don't add CORE constantes after this value!

-- Instance tags
GEM3_INST_TAG_NONE = 0x0000;
 -- TBC
GEM3_INST_TAG_KARA = 0x0001;
GEM3_INST_TAG_ZA = 0x0002;
GEM3_INST_TAG_MAGTH = 0x0004;
GEM3_INST_TAG_GRUUL = 0x0008;
GEM3_INST_TAG_SSC = 0x0010;
GEM3_INST_TAG_EYE = 0x0020;
GEM3_INST_TAG_HYJAL = 0x0040;
GEM3_INST_TAG_BT = 0x0080;
GEM3_INST_TAG_SWP = 0x0100;
 -- WotLK
GEM3_INST_TAG_NAXXRAMAS = 0x0200;
GEM3_INST_TAG_CHAMBEROFASPECTS = 0x0400;
GEM3_INST_TAG_EYEOFETERNITY = 0x0800;
GEM3_INST_TAG_WINTERGRASP = 0x1000;
GEM3_INST_TAG_ULDUAR = 0x2000;
GEM3_INST_TAG_TRIAL_CRUSADER = 0x4000;
GEM3_INST_TAG_ICC = 0x8000;
 -- Classic
GEM3_INST_TAG_MC = 0x02000000;
GEM3_INST_TAG_ZG = 0x04000000;
GEM3_INST_TAG_AQ20 = 0x08000000;
GEM3_INST_TAG_AQ40 = 0x10000000;
GEM3_INST_TAG_NAX = 0x20000000;
GEM3_INST_TAG_BWL = 0x40000000;
GEM3_INST_TAG_ONY = 0x80000000;
GEM3_INST_TAG_UBRS = 0x100000000;

-- Tag Instance type
GEM3_INST_TAG_TYPE_UNKNOWN = 0;
GEM3_INST_TAG_TYPE_RAID_10_NORMAL = 1;
GEM3_INST_TAG_TYPE_RAID_25_NORMAL = 2;
GEM3_INST_TAG_TYPE_RAID_10_HEROIC = 3;
GEM3_INST_TAG_TYPE_RAID_25_HEROIC = 4;

-- GEM3_InstanceTag field access
GEM3_ACCESS_INST_TAG_ID = 1;
GEM3_ACCESS_INST_TAG_RESET = 2;
GEM3_ACCESS_INST_TAG_TYPE = 3;
GEM3_ACCESS_INST_TAG_NAME = 4;

-- Instance type
GEM3_INST_TYPE_NONE = 0;
GEM3_INST_TYPE_BG = 1;
GEM3_INST_TYPE_ARENA = 2;
GEM3_INST_TYPE_PARTY = 3;
GEM3_INST_TYPE_PARTY_HEROIC = 4;
GEM3_INST_TYPE_RAID_10 = 5;
GEM3_INST_TYPE_RAID_10_HEROIC = 6;
GEM3_INST_TYPE_RAID_25 = 7;
GEM3_INST_TYPE_RAID_25_HEROIC = 8;
GEM3_INST_TYPE_RAID_20 = 9; -- Vanilla
GEM3_INST_TYPE_RAID_40 = 10; -- Vanilla
GEM3_INST_TYPE_CUSTOM_START = 100; -- Don't add CORE constantes after this value!

-- Action for incomming event
GEM3_EVT_ACTION_DISCARD        = 0;
GEM3_EVT_ACTION_NEW            = 1;
GEM3_EVT_ACTION_UPDATED        = 2;
GEM3_EVT_ACTION_CRASHED        = 3;
GEM3_EVT_ACTION_RECOVER_UPDATE = 4;

-- Internal error texts (should be in localization file)
GEM3_INTERNAL_ERROR_1 = "Internal Error #1. Please contact Kiki";
GEM3_INTERNAL_ERROR_2 = "Internal Error #2. Please contact Kiki";
GEM3_INTERNAL_ERROR_3 = "Internal Error #3. Please contact Kiki";

-- Channel error codes
GEM3_CHANNEL_ADD_ERROR_NO_ERROR = 0;
GEM3_CHANNEL_ADD_ERROR_NOT_IN_GUILD = 1; -- GEM3_ERR_ADD_CHANNEL_FAILED_NOT_IN_GUILD
GEM3_CHANNEL_ADD_ERROR_MISTAKE_GUILD_NAME = 2; -- GEM3_ERR_ADD_CHANNEL_FAILED_MISTAKE
GEM3_CHANNEL_ADD_ERROR_INVALID_NAME = 3; -- GEM3_ERR_ADD_CHANNEL_FAILED_INVALID
GEM3_CHANNEL_ADD_ERROR_ALREADY_EXISTS = 4; -- GEM3_ERR_ADD_CHANNEL_FAILED_ALREADY_EXISTS
GEM3_CHANNEL_ADD_ERROR_ALIAS_SLASH = 5; -- GEM3_ERR_ADD_CHANNEL_FAILED_ALIAS_SLASH
GEM3_CHANNEL_ADD_ERROR_UNKNOWN = 6; -- GEM3_ERR_ADD_CHANNEL_FAILED_UNKNOWN

GEM3_CHANNEL_UPDT_ERROR_NO_ERROR = 0;
GEM3_CHANNEL_UPDT_ERROR_NOT_FOUND = 1; -- GEM3_ERR_UPDATE_CHANNEL_FAILED_NOT_FOUND
GEM3_CHANNEL_UPDT_ERROR_GUILD_CHANNEL = 2; -- GEM3_ERR_UPDATE_CHANNEL_FAILED_GUILD_CHAN
GEM3_CHANNEL_UPDT_ERROR_ALIAS_SLASH = 3; -- GEM3_ERR_UPDATE_CHANNEL_FAILED_ALIAS_SLASH

-- Zones
  -- LK Outside
  GEM3_ZONE_BOREAN_TUNDRA = 1;
  GEM3_ZONE_CRYSTALSONG_FOREST = 2;
  GEM3_ZONE_DALARAN = 3;
  GEM3_ZONE_DRAGONBLIGHT = 4;
  GEM3_ZONE_GRIZZLY_HILLS = 5;
  GEM3_ZONE_HOWLING_FJORD = 6;
  GEM3_ZONE_ICECROWN = 7;
  GEM3_ZONE_SHOLAZAR_BASIN = 8;
  GEM3_ZONE_THE_STORM_PEAKS = 9;
  GEM3_ZONE_WINTERGRASP = 10;
  GEM3_ZONE_ZUL_DRAK = 11;
  GEM3_ZONE_HROTHGAR_LANDING = 12;
  -- LK Instances
  GEM3_ZONE_ARCHAVON = 21;
  GEM3_ZONE_NAXXRAMAS = 22;
  GEM3_ZONE_CHAMBEROFASPECTS = 23;
  GEM3_ZONE_EYEOFETERNITY = 24;
  GEM3_ZONE_UTGARDE_KEEP = 31;
  GEM3_ZONE_NEXUS = 32;
  GEM3_ZONE_AZJOL_NERUB = 33;
  GEM3_ZONE_AHN_KAHET = 34;
  GEM3_ZONE_DRAK_THARON = 35;
  GEM3_ZONE_GUNDRAK = 36;
  GEM3_ZONE_HALLS_STONE = 37;
  GEM3_ZONE_HALLS_LIGHTNING = 38;
  GEM3_ZONE_UTGARDE_PINNACLE = 39;
  GEM3_ZONE_OCULUS = 40;
  GEM3_ZONE_CULLING_STRATHOLME = 41;
  GEM3_ZONE_VIOLET_HOLD = 42;
  GEM3_ZONE_ULDUAR = 43;
  GEM3_ZONE_TRIAL_CHAMPION = 44; -- Party-5
  GEM3_ZONE_TRIAL_CRUSADER = 45; -- Raid-10/25
  GEM3_ZONE_ICC_FORGE_OF_SOULS = 46; -- Party-5
  GEM3_ZONE_ICC_PIT_OF_SARON = 47; -- Party-5
  GEM3_ZONE_ICC_HALLS_OF_REFLECTION = 48; -- Party-5
  GEM3_ZONE_ICC_RAID = 49; -- Raid-10/25
  -- Capital cities
  GEM3_ZONE_DALARAN = 90;
  GEM3_ZONE_ORGRIMMAR = 91;
  GEM3_ZONE_THUNDERBLUFF = 92;
  GEM3_ZONE_UNDERCITY = 93;
  GEM3_ZONE_SILVERMOON = 94;
  GEM3_ZONE_IRONFORGE = 95;
  GEM3_ZONE_STORMWIND = 96;
  GEM3_ZONE_DARNASSUS = 97;
  GEM3_ZONE_EXODAR = 98;
  GEM3_ZONE_SHATTRATH = 99;

-- Association tables
GEM3_EVTYPE_INSTTYPE_INSTTAG_MATCHING = {
  [GEM3_EVT_TYPE_NONE] = {
    [GEM3_INST_TYPE_NONE] = {GEM3_INST_TAG_NONE},
  },
  [GEM3_EVT_TYPE_TEST] = {
    [GEM3_INST_TYPE_NONE] = {GEM3_INST_TAG_NONE},
    [GEM3_INST_TYPE_BG] = {GEM3_INST_TAG_NONE},
    [GEM3_INST_TYPE_ARENA] = {GEM3_INST_TAG_NONE},
    [GEM3_INST_TYPE_PARTY] = {GEM3_INST_TAG_NONE},
    [GEM3_INST_TYPE_PARTY_HEROIC] = {GEM3_INST_TAG_NONE},
    [GEM3_INST_TYPE_RAID_10] = {GEM3_INST_TAG_NONE,GEM3_INST_TAG_UBRS,GEM3_INST_TAG_KARA,GEM3_INST_TAG_ZA,GEM3_INST_TAG_NAXXRAMAS,GEM3_INST_TAG_CHAMBEROFASPECTS,GEM3_INST_TAG_EYEOFETERNITY,GEM3_INST_TAG_WINTERGRASP,GEM3_INST_TAG_ULDUAR,GEM3_INST_TAG_TRIAL_CRUSADER,GEM3_INST_TAG_ONY,GEM3_INST_TAG_ICC},
    [GEM3_INST_TYPE_RAID_20] = {GEM3_INST_TAG_NONE,GEM3_INST_TAG_ZG,GEM3_INST_TAG_AQ20},
    [GEM3_INST_TYPE_RAID_25] = {GEM3_INST_TAG_NONE,GEM3_INST_TAG_MAGTH,GEM3_INST_TAG_GRUUL,GEM3_INST_TAG_SSC,GEM3_INST_TAG_EYE,GEM3_INST_TAG_HYJAL,GEM3_INST_TAG_BT,GEM3_INST_TAG_SWP,GEM3_INST_TAG_NAXXRAMAS,GEM3_INST_TAG_CHAMBEROFASPECTS,GEM3_INST_TAG_EYEOFETERNITY,GEM3_INST_TAG_WINTERGRASP,GEM3_INST_TAG_ULDUAR,GEM3_INST_TAG_TRIAL_CRUSADER,GEM3_INST_TAG_ONY,GEM3_INST_TAG_ICC},
    [GEM3_INST_TYPE_RAID_40] = {GEM3_INST_TAG_NONE,GEM3_INST_TAG_MC,GEM3_INST_TAG_AQ40,GEM3_INST_TAG_BWL},
    [GEM3_INST_TYPE_RAID_10_HEROIC] = {GEM3_INST_TAG_NONE,GEM3_INST_TAG_TRIAL_CRUSADER,GEM3_INST_TAG_ICC},
    [GEM3_INST_TYPE_RAID_25_HEROIC] = {GEM3_INST_TAG_NONE,GEM3_INST_TAG_TRIAL_CRUSADER,GEM3_INST_TAG_ICC},
  },
  [GEM3_EVT_TYPE_RESET] = {
    [GEM3_INST_TYPE_RAID_10] = {GEM3_INST_TAG_NONE,GEM3_INST_TAG_KARA,GEM3_INST_TAG_ZA,GEM3_INST_TAG_NAXXRAMAS,GEM3_INST_TAG_CHAMBEROFASPECTS,GEM3_INST_TAG_EYEOFETERNITY,GEM3_INST_TAG_WINTERGRASP,GEM3_INST_TAG_ULDUAR,GEM3_INST_TAG_TRIAL_CRUSADER,GEM3_INST_TAG_ONY,GEM3_INST_TAG_ICC},
    [GEM3_INST_TYPE_RAID_20] = {GEM3_INST_TAG_NONE,GEM3_INST_TAG_ZG,GEM3_INST_TAG_AQ20},
    [GEM3_INST_TYPE_RAID_25] = {GEM3_INST_TAG_NONE,GEM3_INST_TAG_MAGTH,GEM3_INST_TAG_GRUUL,GEM3_INST_TAG_SSC,GEM3_INST_TAG_EYE,GEM3_INST_TAG_HYJAL,GEM3_INST_TAG_BT,GEM3_INST_TAG_SWP,GEM3_INST_TAG_NAXXRAMAS,GEM3_INST_TAG_CHAMBEROFASPECTS,GEM3_INST_TAG_EYEOFETERNITY,GEM3_INST_TAG_WINTERGRASP,GEM3_INST_TAG_ULDUAR,GEM3_INST_TAG_TRIAL_CRUSADER,GEM3_INST_TAG_ONY,GEM3_INST_TAG_ICC},
    [GEM3_INST_TYPE_RAID_40] = {GEM3_INST_TAG_NONE,GEM3_INST_TAG_MC,GEM3_INST_TAG_AQ40,GEM3_INST_TAG_BWL},
    [GEM3_INST_TYPE_RAID_10_HEROIC] = {GEM3_INST_TAG_TRIAL_CRUSADER,GEM3_INST_TAG_ICC},
    [GEM3_INST_TYPE_RAID_25_HEROIC] = {GEM3_INST_TAG_TRIAL_CRUSADER,GEM3_INST_TAG_ICC},
  },
  [GEM3_EVT_TYPE_LFG] = {
    [GEM3_INST_TYPE_NONE] = {GEM3_INST_TAG_NONE},
    [GEM3_INST_TYPE_BG] = {GEM3_INST_TAG_NONE},
    [GEM3_INST_TYPE_ARENA] = {GEM3_INST_TAG_NONE},
    [GEM3_INST_TYPE_PARTY] = {GEM3_INST_TAG_NONE},
    [GEM3_INST_TYPE_PARTY_HEROIC] = {GEM3_INST_TAG_NONE},
    [GEM3_INST_TYPE_RAID_10] = {GEM3_INST_TAG_NONE,GEM3_INST_TAG_UBRS,GEM3_INST_TAG_KARA,GEM3_INST_TAG_ZA,GEM3_INST_TAG_NAXXRAMAS,GEM3_INST_TAG_CHAMBEROFASPECTS,GEM3_INST_TAG_EYEOFETERNITY,GEM3_INST_TAG_WINTERGRASP,GEM3_INST_TAG_ULDUAR,GEM3_INST_TAG_TRIAL_CRUSADER,GEM3_INST_TAG_ONY,GEM3_INST_TAG_ICC},
    [GEM3_INST_TYPE_RAID_20] = {GEM3_INST_TAG_NONE,GEM3_INST_TAG_ZG,GEM3_INST_TAG_AQ20},
    [GEM3_INST_TYPE_RAID_25] = {GEM3_INST_TAG_NONE,GEM3_INST_TAG_MAGTH,GEM3_INST_TAG_GRUUL,GEM3_INST_TAG_SSC,GEM3_INST_TAG_EYE,GEM3_INST_TAG_HYJAL,GEM3_INST_TAG_BT,GEM3_INST_TAG_SWP,GEM3_INST_TAG_NAXXRAMAS,GEM3_INST_TAG_CHAMBEROFASPECTS,GEM3_INST_TAG_EYEOFETERNITY,GEM3_INST_TAG_WINTERGRASP,GEM3_INST_TAG_ULDUAR,GEM3_INST_TAG_TRIAL_CRUSADER,GEM3_INST_TAG_ONY,GEM3_INST_TAG_ICC},
    [GEM3_INST_TYPE_RAID_40] = {GEM3_INST_TAG_NONE,GEM3_INST_TAG_MC,GEM3_INST_TAG_AQ40,GEM3_INST_TAG_BWL},
    [GEM3_INST_TYPE_RAID_10_HEROIC] = {GEM3_INST_TAG_NONE,GEM3_INST_TAG_TRIAL_CRUSADER,GEM3_INST_TAG_ICC},
    [GEM3_INST_TYPE_RAID_25_HEROIC] = {GEM3_INST_TAG_NONE,GEM3_INST_TAG_TRIAL_CRUSADER,GEM3_INST_TAG_ICC},
  },
  [GEM3_EVT_TYPE_POLL] = {
    [GEM3_INST_TYPE_NONE] = {GEM3_INST_TAG_NONE},
  },
  [GEM3_EVT_TYPE_PLAYER_AVALABILITY] = {
    [GEM3_INST_TYPE_NONE] = {GEM3_INST_TAG_NONE},
    [GEM3_INST_TYPE_BG] = {GEM3_INST_TAG_NONE},
    [GEM3_INST_TYPE_ARENA] = {GEM3_INST_TAG_NONE},
    [GEM3_INST_TYPE_PARTY] = {GEM3_INST_TAG_NONE},
    [GEM3_INST_TYPE_PARTY_HEROIC] = {GEM3_INST_TAG_NONE},
    [GEM3_INST_TYPE_RAID_10] = {GEM3_INST_TAG_NONE,GEM3_INST_TAG_UBRS,GEM3_INST_TAG_KARA,GEM3_INST_TAG_ZA,GEM3_INST_TAG_NAXXRAMAS,GEM3_INST_TAG_CHAMBEROFASPECTS,GEM3_INST_TAG_EYEOFETERNITY,GEM3_INST_TAG_WINTERGRASP,GEM3_INST_TAG_ULDUAR,GEM3_INST_TAG_TRIAL_CRUSADER,GEM3_INST_TAG_ONY,GEM3_INST_TAG_ICC},
    [GEM3_INST_TYPE_RAID_20] = {GEM3_INST_TAG_NONE,GEM3_INST_TAG_ZG,GEM3_INST_TAG_AQ20},
    [GEM3_INST_TYPE_RAID_25] = {GEM3_INST_TAG_NONE,GEM3_INST_TAG_MAGTH,GEM3_INST_TAG_GRUUL,GEM3_INST_TAG_SSC,GEM3_INST_TAG_EYE,GEM3_INST_TAG_HYJAL,GEM3_INST_TAG_BT,GEM3_INST_TAG_SWP,GEM3_INST_TAG_NAXXRAMAS,GEM3_INST_TAG_CHAMBEROFASPECTS,GEM3_INST_TAG_EYEOFETERNITY,GEM3_INST_TAG_WINTERGRASP,GEM3_INST_TAG_ULDUAR,GEM3_INST_TAG_TRIAL_CRUSADER,GEM3_INST_TAG_ONY,GEM3_INST_TAG_ICC},
    [GEM3_INST_TYPE_RAID_40] = {GEM3_INST_TAG_NONE,GEM3_INST_TAG_MC,GEM3_INST_TAG_AQ40,GEM3_INST_TAG_BWL},
    [GEM3_INST_TYPE_RAID_10_HEROIC] = {GEM3_INST_TAG_NONE,GEM3_INST_TAG_TRIAL_CRUSADER,GEM3_INST_TAG_ICC},
    [GEM3_INST_TYPE_RAID_25_HEROIC] = {GEM3_INST_TAG_NONE,GEM3_INST_TAG_TRIAL_CRUSADER,GEM3_INST_TAG_ICC},
  },
  [GEM3_EVT_TYPE_RAID_AVAILABILITY] = {
    [GEM3_INST_TYPE_NONE] = {GEM3_INST_TAG_NONE},
    [GEM3_INST_TYPE_RAID_10] = {GEM3_INST_TAG_NONE,GEM3_INST_TAG_UBRS,GEM3_INST_TAG_KARA,GEM3_INST_TAG_ZA,GEM3_INST_TAG_NAXXRAMAS,GEM3_INST_TAG_CHAMBEROFASPECTS,GEM3_INST_TAG_EYEOFETERNITY,GEM3_INST_TAG_WINTERGRASP,GEM3_INST_TAG_ULDUAR,GEM3_INST_TAG_TRIAL_CRUSADER,GEM3_INST_TAG_ONY,GEM3_INST_TAG_ICC},
    [GEM3_INST_TYPE_RAID_20] = {GEM3_INST_TAG_NONE,GEM3_INST_TAG_ZG,GEM3_INST_TAG_AQ20},
    [GEM3_INST_TYPE_RAID_25] = {GEM3_INST_TAG_NONE,GEM3_INST_TAG_MAGTH,GEM3_INST_TAG_GRUUL,GEM3_INST_TAG_SSC,GEM3_INST_TAG_EYE,GEM3_INST_TAG_HYJAL,GEM3_INST_TAG_BT,GEM3_INST_TAG_SWP,GEM3_INST_TAG_NAXXRAMAS,GEM3_INST_TAG_CHAMBEROFASPECTS,GEM3_INST_TAG_EYEOFETERNITY,GEM3_INST_TAG_WINTERGRASP,GEM3_INST_TAG_ULDUAR,GEM3_INST_TAG_TRIAL_CRUSADER,GEM3_INST_TAG_ONY,GEM3_INST_TAG_ICC},
    [GEM3_INST_TYPE_RAID_40] = {GEM3_INST_TAG_NONE,GEM3_INST_TAG_MC,GEM3_INST_TAG_AQ40,GEM3_INST_TAG_BWL},
    [GEM3_INST_TYPE_RAID_10_HEROIC] = {GEM3_INST_TAG_NONE,GEM3_INST_TAG_TRIAL_CRUSADER,GEM3_INST_TAG_ICC},
    [GEM3_INST_TYPE_RAID_25_HEROIC] = {GEM3_INST_TAG_NONE,GEM3_INST_TAG_TRIAL_CRUSADER,GEM3_INST_TAG_ICC},

  },
  [GEM3_EVT_TYPE_INSTANCE5] = {
    [GEM3_INST_TYPE_NONE] = {GEM3_INST_TAG_NONE},
    [GEM3_INST_TYPE_PARTY] = {GEM3_INST_TAG_NONE},
  },
  [GEM3_EVT_TYPE_INSTANCE5_BC] = {
    [GEM3_INST_TYPE_NONE] = {GEM3_INST_TAG_NONE},
    [GEM3_INST_TYPE_PARTY] = {GEM3_INST_TAG_NONE},
    [GEM3_INST_TYPE_PARTY_HEROIC] = {GEM3_INST_TAG_NONE},
  },
  [GEM3_EVT_TYPE_INSTANCE5_WOTLK] = {
    [GEM3_INST_TYPE_NONE] = {GEM3_INST_TAG_NONE},
    [GEM3_INST_TYPE_PARTY] = {GEM3_INST_TAG_NONE},
    [GEM3_INST_TYPE_PARTY_HEROIC] = {GEM3_INST_TAG_NONE},
  },
  [GEM3_EVT_TYPE_RAID] = {
    [GEM3_INST_TYPE_NONE] = {GEM3_INST_TAG_NONE},
    [GEM3_INST_TYPE_RAID_10] = {GEM3_INST_TAG_NONE,GEM3_INST_TAG_UBRS},
    [GEM3_INST_TYPE_RAID_20] = {GEM3_INST_TAG_NONE,GEM3_INST_TAG_ZG,GEM3_INST_TAG_AQ20},
    [GEM3_INST_TYPE_RAID_40] = {GEM3_INST_TAG_NONE,GEM3_INST_TAG_MC,GEM3_INST_TAG_AQ40,GEM3_INST_TAG_BWL},
  },
  [GEM3_EVT_TYPE_RAID_BC] = {
    [GEM3_INST_TYPE_RAID_10] = {GEM3_INST_TAG_NONE,GEM3_INST_TAG_KARA,GEM3_INST_TAG_ZA},
    [GEM3_INST_TYPE_RAID_25] = {GEM3_INST_TAG_NONE,GEM3_INST_TAG_MAGTH,GEM3_INST_TAG_GRUUL,GEM3_INST_TAG_SSC,GEM3_INST_TAG_EYE,GEM3_INST_TAG_HYJAL,GEM3_INST_TAG_BT,GEM3_INST_TAG_SWP},
  },
  [GEM3_EVT_TYPE_RAID_WOTLK] = {
    [GEM3_INST_TYPE_RAID_10] = {GEM3_INST_TAG_NONE,GEM3_INST_TAG_NAXXRAMAS,GEM3_INST_TAG_CHAMBEROFASPECTS,GEM3_INST_TAG_EYEOFETERNITY,GEM3_INST_TAG_WINTERGRASP,GEM3_INST_TAG_ULDUAR,GEM3_INST_TAG_TRIAL_CRUSADER,GEM3_INST_TAG_ONY,GEM3_INST_TAG_ICC},
    [GEM3_INST_TYPE_RAID_25] = {GEM3_INST_TAG_NONE,GEM3_INST_TAG_NAXXRAMAS,GEM3_INST_TAG_CHAMBEROFASPECTS,GEM3_INST_TAG_EYEOFETERNITY,GEM3_INST_TAG_WINTERGRASP,GEM3_INST_TAG_ULDUAR,GEM3_INST_TAG_TRIAL_CRUSADER,GEM3_INST_TAG_ONY,GEM3_INST_TAG_ICC},
    [GEM3_INST_TYPE_RAID_10_HEROIC] = {GEM3_INST_TAG_NONE,GEM3_INST_TAG_TRIAL_CRUSADER,GEM3_INST_TAG_ICC},
    [GEM3_INST_TYPE_RAID_25_HEROIC] = {GEM3_INST_TAG_NONE,GEM3_INST_TAG_TRIAL_CRUSADER,GEM3_INST_TAG_ICC},
  },
  [GEM3_EVT_TYPE_PVP] = {
    [GEM3_INST_TYPE_NONE] = {GEM3_INST_TAG_NONE},
    [GEM3_INST_TYPE_BG] = {GEM3_INST_TAG_NONE},
    [GEM3_INST_TYPE_ARENA] = {GEM3_INST_TAG_NONE},
  },
  [GEM3_EVT_TYPE_GUILD_EVENT] = {
    [GEM3_INST_TYPE_NONE] = {GEM3_INST_TAG_NONE},
  },
};

GEM3_INSTTYPE_INSTTAGTYPE_MATCHING = {
  [GEM3_INST_TYPE_RAID_10] = GEM3_INST_TAG_TYPE_RAID_10_NORMAL,
  [GEM3_INST_TYPE_RAID_10_HEROIC] = GEM3_INST_TAG_TYPE_RAID_10_HEROIC,
  [GEM3_INST_TYPE_RAID_25] = GEM3_INST_TAG_TYPE_RAID_25_NORMAL,
  [GEM3_INST_TYPE_RAID_25_HEROIC] = GEM3_INST_TAG_TYPE_RAID_25_HEROIC,
};

GEM3_INSTTAGTYPE_INSTTYPE_MATCHING = {
  [GEM3_INST_TAG_TYPE_RAID_10_NORMAL] = GEM3_INST_TYPE_RAID_10,
  [GEM3_INST_TAG_TYPE_RAID_10_HEROIC] = GEM3_INST_TYPE_RAID_10_HEROIC,
  [GEM3_INST_TAG_TYPE_RAID_25_NORMAL] = GEM3_INST_TYPE_RAID_25,
  [GEM3_INST_TAG_TYPE_RAID_25_HEROIC] = GEM3_INST_TYPE_RAID_25_HEROIC,
  [GEM3_INST_TAG_TYPE_UNKNOWN] = GEM3_INST_TYPE_NONE,
};
