--[[
  Custom Chat Comm, by Kiki - Conseil des Ombres-EU
  

  TODO:
   - Detecter que 10 chans sont deja joint
   - [Detecter un mauvais mot de passe]
   - Tester qd on relog a griphon

 ChangeLog:
   - 2009/09/15 : 1.2
     - Increased delay before auto rejoin of a manually left channel
   - 2008/12/15 : 1.1
     - Fixed typo in UnregisterAlias and UnregisterSlash prototypes
   - 2008/11/26 : 1.0
     - Added the function 'CustomChatComm.CheckNameValidity(channelName)' to check if a channel is valid for CCC
   - 2008/11/20 : 0.9
     - Improved channel init (to avoid having a channel in /1 instead of General or Trade)
     - Automatically setting new channel visible in DEFAULT_CHAT_FRAME
   - 2008/11/06 : 0.8
     - Added support for alias and slash
   - 2008/09/19 : 0.7
     - Checking for invalid channel name
   - 2008/09/05 : 0.6
     - Fixed incorrect ChatThrottleLib queuing for last section of a multiple parts message
   - 2008/09/04 : 0.5
     - Fixed UTF-8 double/triple bytes chars splitting
   - 2008/09/03 : 0.4
     - Fixed UnregisterComm function not defined in metatable
   - 2008/08/25 : 0.3
     - Fixed password initialization if alone in the channel
     - Fixed channel initialization detection
   - 2008/06/23 : 0.2
     - Startup don't init channel if unit is on taxi
   - 2008/06/18 : 0.1
     - Library Created
  
  Credits:
   - Message splitting code grabbed from AceComm
]]

local CCC_VERS = 1.2;
local CCC_DEBUG = false;

--------------- Quick access ---------------

local type = type
local strsub = string.sub
local strgsub = string.gsub;
local strfind = string.find
local tinsert = table.insert
local tconcat = table.concat
local strlower = string.lower;
local strupper = string.upper;

--------------- Version Check ---------------

local isBetterInstanceLoaded = ( CustomChatComm and CustomChatComm.version and CustomChatComm.version >= CCC_VERS );

if (not isBetterInstanceLoaded) then

if(not CustomChatComm)
then
  CustomChatComm = {};
  CustomChatComm.comms = {};
end
CustomChatComm.version = CCC_VERS;

local CallbackHandler = LibStub:GetLibrary("CallbackHandler-1.0")

--------------- AceComm-3.0,4 grabbed vars ---------------

-- for my sanity and yours, let's give the message type bytes names
local MSG_MULTI_FIRST = "\001"
local MSG_MULTI_NEXT  = "\002"
local MSG_MULTI_LAST  = "\003"

CustomChatComm.multipart_origprefixes = CustomChatComm.multipart_origprefixes or {} -- e.g. "Prefix\001"="Prefix", "Prefix\002"="Prefix"
CustomChatComm.multipart_reassemblers = CustomChatComm.multipart_reassemblers or {} -- e.g. "Prefix\001"="OnReceiveMultipartFirst"

-- the multipart message spool: indexed by a combination of sender+distribution+
CustomChatComm.multipart_spool = CustomChatComm.multipart_spool or {} 


--------------- Shared Constantes ---------------
--------------- Shared variables ---------------
--------------- Local Constantes ---------------

local CCC_CB_YOU_JOINED = "YOU_JOINED";
local CCC_CB_YOU_LEFT = "YOU_LEFT";

--------------- Local variables ---------------

local _CCC_init_timer = nil;
local _CCC_init_ready = false;
local _CCC_blizz_channel_joined = false;
local _CCC_LogMainTable = nil;

--------------- Internal functions ---------------

local function _CCC_Log(buffer,msg)
  if(_CCC_LogMainTable == nil) then return end
  if(_CCC_LogMainTable[buffer] == nil) then _CCC_LogMainTable[buffer] = {} end
  tinsert(_CCC_LogMainTable[buffer],msg);
end

local function _CCC_WoWAPI_FireCallback(event,...)
  local channel = strlower(select(9,...));

  for i,com in ipairs(CustomChatComm.comms)
  do
    if(com.channel == channel)
    then
      local cb = com.callbacks[event];
      if(cb)
      then
        cb(...);
      end
    end
  end
end

local function _CCC_Channel_FireCallback(event,channel,...)
  for i,com in ipairs(CustomChatComm.comms)
  do
    if(com.channel == channel)
    then
      local cb = com.callbacks[event];
      if(cb)
      then
        cb(channel,...);
      end
    end
  end
end

local function _CCC_InitComm(com)
  if(GetChannelName(com.channel) ~= 0)
  then
    --CustomChatComm.ChatDebug("chan is already init: "..com.channel);
    com.isInit = true;
    ListChannelByName(com.channel);
    _CCC_Channel_FireCallback(CCC_CB_YOU_JOINED,com.channel);
    return;
  end

  --CustomChatComm.ChatDebug("Joining chan "..com.channel);
  JoinChannelByName(com.channel,com.password);
  if(DEFAULT_CHAT_FRAME) -- Enforce channel visibility
  then
    ChatFrame_AddChannel(DEFAULT_CHAT_FRAME,com.channel);
  end
end

local function _CCC_EnforceRejoin(com)
  if(com)
  then
    JoinChannelByName(com.channel,com.password);
  end
end

local function _CCC_EnforcePassword(channel)
  for i,com in ipairs(CustomChatComm.comms)
  do
    if(channel == com.channel)
    then
      if(com.password and com.password ~= "")
      then
        SetChannelPassword(channel,com.password);
      end
    end
  end
end

local function _CCC_AlreadyRegisteredChannel(channel)
  for i,com in ipairs(CustomChatComm.comms)
  do
    if(com.channel == channel)
    then
      return true;
    end
  end
  return false;
end

local function _CCC_GetCom(channel)
  if(channel == nil)
  then
    return nil;
  end
  for i,com in ipairs(CustomChatComm.comms)
  do
    if(com.channel == channel)
    then
      return com;
    end
  end
  return nil;
end

local function _CCC_ProcessChannelListing(event,listing, _, _, _, _, _, _, _, channelName)
  local channel = strlower(channelName);
  local cLeader = strbyte("*");
  local cModerator = strbyte("@");
  local cMute = strbyte("#");
  local char1;
  local playerName = UnitName("player");
  
  for value in gmatch(listing,"[^, ]+")
  do
    local isLeader = false;
    char1 = strbyte(value);
    if(char1 == cLeader)
    then
      value = strsub(value,2);
      if(value == playerName)
      then
        isLeader = true;
      end
    elseif(char1 == cModerator or char1 == cMute)
    then
      value = strsub(value,2);
    end
    if(isLeader)
    then
      _CCC_EnforcePassword(channel);
    end
  end
end

local function _CCC_ProcessChannelNotice(event,kind,player1,_,_,_,_,_,_,channelName)
  local channel = strlower(channelName);

  if(kind == "YOU_JOINED")
  then
    if(_CCC_AlreadyRegisteredChannel(channel))
    then
      ListChannelByName(channel);
    end
    _CCC_Channel_FireCallback(CCC_CB_YOU_JOINED,channel);
  elseif(kind == "YOU_LEFT")
  then
    _CCC_Channel_FireCallback(CCC_CB_YOU_LEFT,channel);
    -- Enforce rejoin
    local done = false;
    for _,com in ipairs(CustomChatComm.comms)
    do
      if(channel == com.channel)
      then
        com.isInit = false;
        if(done == false)
        then
          Chronos.schedule(5,_CCC_EnforceRejoin,com);
          done = true;
        end
      end
    end
  elseif(kind == "OWNER_CHANGED")
  then
    if(player1 == UnitName("player"))
    then
      _CCC_EnforcePassword(channel);
    end
  elseif(kind == "WRONG_PASSWORD")
  then
  end
end

local function _CCC_InitComplete()
  for i,com in ipairs(CustomChatComm.comms)
  do
    if(com.isInit == false)
    then
      _CCC_InitComm(com);
    end
  end
end

local function _CCC_CheckPlayerGuild()
  if(IsInGuild()) -- Have a guild
  then
    if(GetGuildInfo("player") == nil) -- Guild name still not init
    then
      return false;
    end
  end
  return true;
end

local function _CCC_ProcessChannelMessage(channel,sender,text)
  if(_CCC_AlreadyRegisteredChannel(channel) == false) -- Not for me
  then
    return;
  end
  if(CCC_DEBUG) then _CCC_Log("recv",text) end
  -- Try to get prefix and message
  local _,_,prefix,message = strfind(text,"([^\007]+)\007(.+)");
  -- Following code grabbed from AceComm-3.0,4
  local reassemblername = CustomChatComm.multipart_reassemblers[prefix]
  if reassemblername then
    -- multipart: reassemble
    local CustomChatCommReassemblerFunc = CustomChatComm[reassemblername]
    local origprefix = CustomChatComm.multipart_origprefixes[prefix]
    CustomChatCommReassemblerFunc(CustomChatComm, origprefix, message, channel, sender)
  else
    -- single part: fire it off immediately and let CallbackHandler decide if it's registered or not
    CustomChatComm.callbacks:Fire(prefix, message, channel, sender)
  end
end

local function _CCC_NameFromSlash(slashName)
  local slash = strupper(slashName);

  for i,com in ipairs(CustomChatComm.comms)
  do
    if(com.slash and com.slash == slash)
    then
      return GetChannelName(com.channel);
    end
  end
  return nil;
end

local _CCC_PIPE_ENTITIE = "\127p";
local function _CCC_ConvertLink(msg)
  return strgsub(msg,"|",_CCC_PIPE_ENTITIE);
  --return msg;
end

local function _CCC_UnconvertLink(msg)
  return strgsub(msg,_CCC_PIPE_ENTITIE,"|");
  --return msg;
end

local old_SendChatMessage = SendChatMessage;
local function _CCC_SendChatMessage(msg,sys,lang,name)
  local ret;
  local afk = GetCVar("autoClearAFK");
  SetCVar("autoClearAFK", 0);

  local id = _CCC_NameFromSlash(sys);
  if(id)
  then
    ret = old_SendChatMessage(_CCC_ConvertLink(msg),"CHANNEL",lang,id);
  else
    ret = old_SendChatMessage(msg,sys,lang,name);
  end
  SetCVar("autoClearAFK",afk);
  return ret;
end
SendChatMessage = _CCC_SendChatMessage;

local old_ChatFrame_MessageEventHandler = ChatFrame_MessageEventHandler;
local function _CCC_ChatFrame_MessageEventHandler(self,event,...)
  local channelName = select(9,...);
  local channel = channelName and strlower(channelName);
  local com = _CCC_GetCom(channel);
  if(com) -- For me
  then
    if(event == "CHAT_MSG_CHANNEL")
    then
      local arg1,arg2,arg3,arg4,arg5,arg6,arg7,arg8,arg9,arg10,arg11,arg12,arg13,arg14,arg15 = ...;

      if(strfind(arg1,"^GEM3")) -- A GEM message ? Don't display
      then
        return true; -- Filtered
      end
      -- the message is shown in this ChatFrame ?
      local info;
      local found = 0;
      local channelLength = strlen(arg4);
      for index, value in pairs(self.channelList) do
        if ( channelLength > strlen(value) ) then
          -- arg9 is the channel name without the number in front...
          if ( ((arg7 > 0) and (self.zoneChannelList[index] == arg7)) or (strlower(value) == channel) ) then
            found = 1;
            info = ChatTypeInfo["CHANNEL"..arg8];
            break;
          end
        end
      end
      if((found == 1) and info)
      then
        -- Convert chat link
        arg1 = _CCC_UnconvertLink(arg1);
        -- Convert alias
        if(com.slash)
        then
          event = "CHAT_MSG_"..com.slash;
          arg4 = "";
        end
        return old_ChatFrame_MessageEventHandler(self,event,arg1,arg2,arg3,arg4,arg5,arg6,arg7,arg8,arg9,arg10,arg11,arg12,arg13,arg14,arg15);
      end
    end
  end
  return old_ChatFrame_MessageEventHandler(self,event,...);
end
ChatFrame_MessageEventHandler = _CCC_ChatFrame_MessageEventHandler;

----------------------------------------
-- Message receiving -- Grabbed from AceComm-3.0,4
----------------------------------------

do
	local compost = setmetatable({}, {__mode=="k"})
	local function new()
		local t = next(compost)
		if t then 
			compost[t]=nil
			for i=#t,3,-1 do	-- faster than pairs loop. don't even nil out 1/2 since they'll be overwritten
				t[i]=nil
			end
			return t
		end
		
		return {}
	end
	
	local function lostdatawarning(prefix,sender,where)
		DEFAULT_CHAT_FRAME:AddMessage(MAJOR..": Warning: lost network data regarding '"..tostring(prefix).."' from '"..tostring(sender).."' (in "..where..")")
	end

	function CustomChatComm:OnReceiveMultipartFirst(prefix, message, distribution, sender)
		local key = prefix.."\t"..distribution.."\t"..sender	-- a unique stream is defined by the prefix + distribution + sender
		local spool = CustomChatComm.multipart_spool
		
		--[[
		if spool[key] then 
			lostdatawarning(prefix,sender,"First")
			-- continue and overwrite
		end
		--]]
		
		spool[key] = message	-- plain string for now
	end

	function CustomChatComm:OnReceiveMultipartNext(prefix, message, distribution, sender)
		local key = prefix.."\t"..distribution.."\t"..sender	-- a unique stream is defined by the prefix + distribution + sender
		local spool = CustomChatComm.multipart_spool
		local olddata = spool[key]
		
		if not olddata then
			--lostdatawarning(prefix,sender,"Next")
			return
		end

		if type(olddata)~="table" then
			-- ... but what we have is not a table. So make it one. (Pull a composted one if available)
			local t = new()
			t[1] = olddata		-- add old data as first string
			t[2] = message      -- and new message as second string
			spool[key] = t		-- and put the table in the spool instead of the old string
		else
			tinsert(olddata, message)
		end
	end

	function CustomChatComm:OnReceiveMultipartLast(prefix, message, distribution, sender)
		local key = prefix.."\t"..distribution.."\t"..sender	-- a unique stream is defined by the prefix + distribution + sender
		local spool = CustomChatComm.multipart_spool
		local olddata = spool[key]
		
		if not olddata then
			--lostdatawarning(prefix,sender,"End")
			return
		end

		spool[key] = nil
		
		if type(olddata)=="table" then
			-- if we've received a "next", the spooled data will be a table for rapid & garbage-free tconcat
			tinsert(olddata, message)
			CustomChatComm.callbacks:Fire(prefix, tconcat(olddata, ""), distribution, sender)
			compost[olddata]=true
		else
			-- if we've only received a "first", the spooled data will still only be a string
			CustomChatComm.callbacks:Fire(prefix, olddata..message, distribution, sender)
		end
	end
end

--------------- Shared functions ---------------

function CustomChatComm.Todo(str,...)
  if(DEFAULT_CHAT_FRAME)
  then
    DEFAULT_CHAT_FRAME:AddMessage("CustomChatComm TODO: "..string.format(str,...), 1.0, 0.7, 0.15);
  end
end

function CustomChatComm.ChatDebug(str)
  if(DEFAULT_CHAT_FRAME)
  then
    DEFAULT_CHAT_FRAME:AddMessage("CustomChatComm: "..str, 1.0, 0, 0);
  end
end

function CustomChatComm.ChatPrint(str)
  if(DEFAULT_CHAT_FRAME)
  then
    DEFAULT_CHAT_FRAME:AddMessage("CustomChatComm: "..str, 1.0, 0.7, 0.15);
  end
end

function CustomChatComm.CheckNameValidity(channelName)
  -- Sanity check for channelName
  local _,_,name = strfind(channelName,"^([^%d]).+") -- Check for starting number
  if(name == nil)
  then
    return false;
  end
  local _,_,name = strfind(channelName,"^.*([%c%p%s]).*$") -- Check for invalid char
  if(name ~= nil)
  then
    return false;
  end
  return true;
end

function CustomChatComm.RegisterComm(prefix,channelName,password,callback_message,callback_you_joined,callback_you_left)
  if(CustomChatComm.CheckNameValidity(channelName) == false)
  then
    return nil;
  end

  local com = {};

  com.prefix = prefix;
  com.channel = strlower(channelName);
  com.password = password;
  com.callbacks = {};
  com.callbacks[CCC_CB_YOU_JOINED] = callback_you_joined;
  com.callbacks[CCC_CB_YOU_LEFT] = callback_you_left;
  com.isInit = false;
  CustomChatComm._RegisterComm(CustomChatComm,prefix,callback_message) -- created by CallbackHandler

  -- Add new com to list
  tinsert(CustomChatComm.comms,com); -- Add before checking for Init, or cb won't fire if channel already joined

  -- If global channels init is complete, join now
  -- Or if channel already joined (reloadui), complete init
  if(_CCC_init_ready or GetChannelName(com.channel) ~= 0)
  then
    _CCC_InitComm(com);
  end
  
  return com;
end

function CustomChatComm.UnregisterComm(prefix,channelName)
  local channel = strlower(channelName);

  for i,com in ipairs(CustomChatComm.comms)
  do
    if(prefix == com.prefix and channel == com.channel)
    then
      tremove(CustomChatComm.comms,i);
      if(_CCC_AlreadyRegisteredChannel(channel) == false) -- Last one
      then
        LeaveChannelByName(channel);
      end
    end
  end
end

local function _CCC_dummyfunction()
end

function CustomChatComm.RegisterSlash(prefix,channelName,slash)
  local channel = strlower(channelName);

  for i,com in ipairs(CustomChatComm.comms)
  do
    if(prefix == com.prefix and channel == com.channel)
    then
      local upper = strupper(slash);
      local id = GetChannelName(channel);
      --SlashCmdList[upper] = _CCC_dummyfunction;
      setglobal("SLASH_"..upper.."1", "/"..slash);
      ChatTypeInfo["CHANNEL"..id].sticky = 1;
      ChatTypeInfo[upper] = ChatTypeInfo["CHANNEL"..id];
      local alias = id..". "..channelName;
      if(com.alias)
      then
        alias = com.alias;
        setglobal("CHAT_MSG_"..upper,alias);
      end
      setglobal("CHAT_"..upper.."_GET", "["..alias.."] %s:\32");
      setglobal("CHAT_"..upper.."_SEND", alias..":\32");
      com.slash = strupper(slash);
    end
  end
end

function CustomChatComm.UnregisterSlash(prefix,channelName)
  local channel = strlower(channelName);

  for i,com in ipairs(CustomChatComm.comms)
  do
    if(prefix == com.prefix and channel == com.channel and com.slash)
    then
      local upper = com.slash;
      SlashCmdList[upper] = nil;
      setglobal("SLASH_"..upper.."1", nil);
      setglobal("CHAT_MSG_"..upper, nil);
      setglobal("CHAT_"..upper.."_GET", nil);
      setglobal("CHAT_"..upper.."_SEND", nil);
      com.slash = nil;
    end
  end
end

function CustomChatComm.RegisterAlias(prefix,channelName,alias)
  local channel = strlower(channelName);

  for i,com in ipairs(CustomChatComm.comms)
  do
    if(prefix == com.prefix and channel == com.channel)
    then
      if(com.slash)
      then
        local upper = com.slash;
        local id = GetChannelName(channel);
        setglobal("CHAT_MSG_"..upper,alias);
        setglobal("CHAT_"..upper.."_GET", "["..alias.."] %s:\32");
        setglobal("CHAT_"..upper.."_SEND", alias..":\32");
      end
      com.alias = alias;
    end
  end
end

function CustomChatComm.UnregisterAlias(prefix,channelName)
  local channel = strlower(channelName);

  for i,com in ipairs(CustomChatComm.comms)
  do
    if(prefix == com.prefix and channel == com.channel and com.alias)
    then
      if(com.slash)
      then
        local upper = com.slash;
        local id = GetChannelName(channel);
        local alias = id..". "..channelName;
        setglobal("CHAT_MSG_"..upper,nil);
        setglobal("CHAT_"..upper.."_GET", "["..alias.."] %s:\32");
        setglobal("CHAT_"..upper.."_SEND", alias..":\32");
      end
      com.alias = nil;
    end
  end
end

function CustomChatComm.RegisterCallback(com,callback_name,callback)
  com.callbacks[callback_name] = callback;
end

function CustomChatComm.UnregisterCallback(com,callback_name)
  com.callbacks[callback_name] = nil;
end

local UTF2CHAR = {
  [0xc2] = true,
  [0xc3] = true,
  [0xce] = true,
}

local UTF3CHAR = {
  [0xe2] = true,
}

local function _CCC_strsub(text,pos1,pos2) -- Special function to avoid UTF double/triple chars splitting (causes a 'Chat text must be in UTF-8 format' error message)
  local len = pos2 - pos1;
  local val = text:byte(pos2);
  if(UTF2CHAR[val]) -- UTF-8 2-bytes character
  then
    len = len - 1;
  end
  if(UTF3CHAR[val]) -- UTF-8 3-bytes character
  then
    len = len - 2;
  end
  return strsub(text,pos1,pos1+len),len+1;
end

function CustomChatComm.SendChatMessage(prefix,text,channelName)
  local index = GetChannelName(channelName);
  local channel = strlower(channelName);
  if(index and index > 0)
  then
    -- Following code grabbed from AceComm-3.0,4 -- Modified by Kiki to avoid UTF-8 chars splitting
    if strfind(prefix, "[\001-\003]") then
      error("CustomChatComm.SendChatMessage: Characters \\001--\\003) in prefix are reserved", 2);
    end

    local textlen = #text;
    local maxtextlen = 252 - #prefix; -- 252 is the max length of prefix + text that can be sent in one message <- Let's use 252 for more safety
    local queueName = prefix..channel;

    if textlen <= maxtextlen then
      -- fits all in one message
      if(CCC_DEBUG) then _CCC_Log("send",prefix.."\007"..text) end
      ChatThrottleLib:SendChatMessage("NORMAL",prefix,prefix.."\007"..text,"CHANNEL",nil,index,queueName);
    else
      maxtextlen = maxtextlen - 1	-- 1 extra byte for part indicator in suffix
      -- first part
      local chunk,chunklen = _CCC_strsub(text, 1, maxtextlen)
      if(CCC_DEBUG) then _CCC_Log("send",prefix..MSG_MULTI_FIRST.."\007"..chunk) end
      ChatThrottleLib:SendChatMessage("NORMAL",prefix,prefix..MSG_MULTI_FIRST.."\007"..chunk,"CHANNEL",nil,index,queueName);
      -- continuation
      local pos = chunklen + 1
      local prefix2 = prefix..MSG_MULTI_NEXT

      while pos+maxtextlen <= textlen do
        chunk,chunklen = _CCC_strsub(text, pos, pos+maxtextlen-1)
        if(CCC_DEBUG) then _CCC_Log("send",prefix2.."\007"..chunk) end
        ChatThrottleLib:SendChatMessage("NORMAL",prefix,prefix2.."\007"..chunk,"CHANNEL",nil,index,queueName);
        pos = pos + chunklen
      end

      -- final part
      chunk = strsub(text, pos)
      if(CCC_DEBUG) then _CCC_Log("send",prefix..MSG_MULTI_LAST.."\007"..chunk) end
      ChatThrottleLib:SendChatMessage("NORMAL",prefix,prefix..MSG_MULTI_LAST.."\007"..chunk,"CHANNEL",nil,index,queueName);
    end
    
    return true;
  end

  return false;
end

function CustomChatComm.SetDebugMode(mode)
  CCC_DEBUG = mode;
end

function CustomChatComm.SetLogTable(myTable)
  _CCC_LogMainTable = myTable;
end

function CustomChatComm.ResetLogTable()
  if(_CCC_LogMainTable)
  then
    _CCC_LogMainTable.send = nil;
    _CCC_LogMainTable.recv = nil;
  end
end

function CustomChatComm.TestSend(channel)
  local index = GetChannelName(channel);
  if(_CCC_LogMainTable and _CCC_LogMainTable.send and index)
  then
    for i,str in ipairs(_CCC_LogMainTable.send)
    do
      DEFAULT_CHAT_FRAME:AddMessage("Sending line "..i.." of "..#_CCC_LogMainTable.send, 0.9, 0.8, 0.10);
      ChatThrottleLib:SendChatMessage("NORMAL","CCCDebug",str,"CHANNEL",nil,index,"CCCDebug");
    end
  end
end


--------------- Event function ---------------

function CustomChatComm.OnUpdate(self,dt)
  if(_CCC_init_timer == nil) -- Variables not loaded yet
  then
    return;
  end
  _CCC_init_timer = _CCC_init_timer + dt;
  
  if(_CCC_init_timer > 4.0) -- Time to check for full init (after 4 sec)
  then
    local playerName = UnitName("player");
    if(playerName == nil or playerName == UNKNOWNBEING or playerName == UKNOWNBEING or playerName == UNKNOWNOBJECT or _CCC_CheckPlayerGuild() == false or UnitOnTaxi("player") == 1) -- Not ready yet
    then
      return;
    end
    if(_CCC_blizz_channel_joined == false) -- Blizz channel not joined yet
    then
      if(_CCC_init_timer < 30.0) -- Give 30 sec to load a blizz channel, otherwise we assume the player don't have any
      then
        return;
      end
      --CustomChatComm.ChatDebug("Player should not have any blizz channel, go on with channel init");
    end
    _CCC_init_ready = true;
  end
  if(_CCC_init_ready)
  then
    _CCC_InitComplete();
    self:SetScript("OnUpdate",nil);
  end
end

function CustomChatComm.OnEvent(self,event,...)
  if(event == "CHAT_MSG_CHANNEL")
  then
    local text, sender, _, _, _, _, _, _, channel = select(1,...);
    _CCC_ProcessChannelMessage(strlower(channel),sender,text);
  elseif(event == "CHAT_MSG_CHANNEL_JOIN")
  then
    _CCC_WoWAPI_FireCallback(event,...);
  elseif(event == "CHAT_MSG_CHANNEL_LEAVE")
  then
    _CCC_WoWAPI_FireCallback(event,...);
  elseif(event == "CHAT_MSG_CHANNEL_LIST")
  then
    _CCC_ProcessChannelListing(event,...);
    _CCC_WoWAPI_FireCallback(event,...);
  elseif(event == "CHAT_MSG_CHANNEL_NOTICE")
  then
    local msg_type,_,_,_,_,_,chan_type = ...;
    if((chan_type == 1 or chan_type == 2) and msg_type == "YOU_JOINED") -- Joined a global blizzard channel
    then
      _CCC_blizz_channel_joined = true;
      --CustomChatComm.ChatDebug("Joined a blizz global channel");
    end
    _CCC_ProcessChannelNotice(event,...);
    _CCC_WoWAPI_FireCallback(event,...);
  elseif(event == "CHAT_MSG_CHANNEL_NOTICE_USER")
  then
    _CCC_ProcessChannelNotice(event,...);
    _CCC_WoWAPI_FireCallback(event,...);
  elseif(event == "VARIABLES_LOADED")
  then
    _CCC_init_timer = 0; -- Start timer now
  end
end

----------------------------------------
-- Embed CallbackHandler -- Grabbed from AceComm-3.0,4
----------------------------------------

if not CustomChatComm.callbacks then
	-- ensure that 'prefix to watch' table is consistent with registered
	-- callbacks
	CustomChatComm.__prefixes = {}

	CustomChatComm.callbacks = CallbackHandler:New(CustomChatComm,
						"_RegisterComm",
						"_UnregisterComm",
						"UnregisterAllComm")
end

function CustomChatComm.callbacks:OnUsed(target, prefix)
	CustomChatComm.multipart_origprefixes[prefix..MSG_MULTI_FIRST] = prefix
	CustomChatComm.multipart_reassemblers[prefix..MSG_MULTI_FIRST] = "OnReceiveMultipartFirst"
	
	CustomChatComm.multipart_origprefixes[prefix..MSG_MULTI_NEXT] = prefix
	CustomChatComm.multipart_reassemblers[prefix..MSG_MULTI_NEXT] = "OnReceiveMultipartNext"
	
	CustomChatComm.multipart_origprefixes[prefix..MSG_MULTI_LAST] = prefix
	CustomChatComm.multipart_reassemblers[prefix..MSG_MULTI_LAST] = "OnReceiveMultipartLast"
end

function CustomChatComm.callbacks:OnUnused(target, prefix)
	CustomChatComm.multipart_origprefixes[prefix..MSG_MULTI_FIRST] = nil
	CustomChatComm.multipart_reassemblers[prefix..MSG_MULTI_FIRST] = nil
	
	CustomChatComm.multipart_origprefixes[prefix..MSG_MULTI_NEXT] = nil
	CustomChatComm.multipart_reassemblers[prefix..MSG_MULTI_NEXT] = nil
	
	CustomChatComm.multipart_origprefixes[prefix..MSG_MULTI_LAST] = nil
	CustomChatComm.multipart_reassemblers[prefix..MSG_MULTI_LAST] = nil
end


--------------- CustomChatComm Main Frame ---------------

  --Event Driver
  if(not CustomChatCommFrame)
  then
    CreateFrame("Frame", "CustomChatCommFrame");
  end
  --Event Registration
  CustomChatCommFrame:RegisterEvent("VARIABLES_LOADED");
  CustomChatCommFrame:RegisterEvent("CHAT_MSG_CHANNEL");
  CustomChatCommFrame:RegisterEvent("CHAT_MSG_CHANNEL_JOIN");
  CustomChatCommFrame:RegisterEvent("CHAT_MSG_CHANNEL_LEAVE");
  CustomChatCommFrame:RegisterEvent("CHAT_MSG_CHANNEL_LIST");
  CustomChatCommFrame:RegisterEvent("CHAT_MSG_CHANNEL_NOTICE");
  CustomChatCommFrame:RegisterEvent("CHAT_MSG_CHANNEL_NOTICE_USER");
  
  --Frame Scripts
  CustomChatCommFrame:SetScript("OnEvent", CustomChatComm.OnEvent);
  CustomChatCommFrame:SetScript("OnUpdate",CustomChatComm.OnUpdate);
  -- Print init message
  CustomChatComm.ChatPrint("Version "..CCC_VERS.." Loaded !");

end -- not isBetterInstanceLoaded

