--[[
  Guild Event Manager (GEM) - Version 3
  by Kiki from European "Conseil des Ombres" (Horde)
  ----------------------------------------
  Sorting plugin - Uses subscription TimeStamp as priority, and Minimum limits, and guild Ranks
]]

local GEM3_SORT_NAME = nil;
local GEM3_SORT_HELP = nil;
if ( GetLocale() == "frFR" ) then
GEM3_SORT_NAME = "Rang de guilde";
GEM3_SORT_HELP = "Tri les joueurs en fonction de leur rang de guilde puis de leur date d'inscription en respectant les minimas de classe/role.";
--[[elseif ( GetLocale() == "deDE" ) then
GEM3_SORT_NAME = "Datum";
GEM3_SORT_HELP = "Sortiert Spieler unter Beachtung ihrer Anmeldezeit.";]]
else
GEM3_SORT_NAME = "Guild ranks";
GEM3_SORT_HELP = "Sort players using their guild ranks then subscription time, filling minimums first.";
end

-- Quick acces
local tinsert = table.insert;
local tremove = table.remove;
local tsort = table.sort;

--------------- Internal functions ---------------

local function _GEM3_SORT_DpsRoomForMe(event,player)
  if(player.role == GEM3_ROLE_CDPS or player.role == GEM3_ROLE_RDPS)
  then
    local dps_lim = event.roles[GEM3_ROLE_DPS];
    local count = event.roles[GEM3_ROLE_CDPS].sort_stamp_count + event.roles[GEM3_ROLE_RDPS].sort_stamp_count;
    if(dps_lim.max and count > dps_lim.max) -- Dps limit reached
    then
      return false;
    end
  end
  return true;
end

local function _GEM3_SORT_CheckGlobalLimit(event,player,count)
  local clas_lim = event.classes[player.class];
  local rol_lim = event.roles[player.role];

  -- Check if exceeding class limit
  if(count > event.max_count) -- Global limit reached
  then
    if((clas_lim.min and clas_lim.sort_stamp_count <= clas_lim.min)
       or (rol_lim and rol_lim.min and rol_lim.sort_stamp_count <= rol_lim.min))
    then
      GEM3_ChatDebug(GEM3_DEBUG_SORT,"Plugin Stamp: Global limit reached but minima fails if I remove this player, keeping him as titular");
      return true;
    end
    GEM3_ChatDebug(GEM3_DEBUG_SORT,"Plugin Stamp: Global limit reached for "..player.name.." ("..count.." > "..event.max_count..")");
    return false;
  end
  return true; -- Check is OK
end

local function _GEM3_SORT_CheckRoleClassLimits(event,player,count)
  local clas_lim = event.classes[player.class];
  local rol_lim = event.roles[player.role];

  -- Check if exceeding class limit
  if(clas_lim.max and clas_lim.sort_stamp_count > clas_lim.max) -- Class limit reached
  then
    GEM3_ChatDebug(GEM3_DEBUG_SORT,"Plugin Stamp: Class limit reached for "..player.name..", check minimas ("..clas_lim.sort_stamp_count.." > "..tostring(clas_lim.min)..")");
    if(clas_lim.min and clas_lim.sort_stamp_count <= clas_lim.min)
    then
      GEM3_ChatDebug(GEM3_DEBUG_SORT,"Plugin Stamp: Class minima fails if I remove this player, keeping him as titular");
      return true;
    end
    if(rol_lim and rol_lim.min and rol_lim.sort_stamp_count <= rol_lim.min)
    then
      GEM3_ChatDebug(GEM3_DEBUG_SORT,"Plugin Stamp: Role minima fails if I remove this player, keeping him as titular");
      return true;
    end
    return false;
  elseif(rol_lim and rol_lim.max and rol_lim.sort_stamp_count > rol_lim.max) -- Role limit reached
  then
    GEM3_ChatDebug(GEM3_DEBUG_SORT,"Plugin Stamp: Role limit reached for "..player.name..", check minimas ("..rol_lim.sort_stamp_count.." > "..tostring(rol_lim.min)..")");
    if(rol_lim.min and rol_lim.sort_stamp_count <= rol_lim.min)
    then
      GEM3_ChatDebug(GEM3_DEBUG_SORT,"Plugin Stamp: Role minima fails if I remove this player, keeping him as titular");
      return true;
    end
    if(clas_lim.min and clas_lim.sort_stamp_count <= clas_lim.min)
    then
      GEM3_ChatDebug(GEM3_DEBUG_SORT,"Plugin Stamp: Class minima fails if I remove this player, keeping him as titular");
      return true;
    end
    return false;
  elseif(not _GEM3_SORT_DpsRoomForMe(event,player))
  then
    GEM3_ChatDebug(GEM3_DEBUG_SORT,"Plugin Stamp: DPS limit reached for "..player.name);
    return false;
  end
  return true; -- Check is OK
end

local function _GEM3_SORT_CheckClassLimits(event,player,count)
  local clas_lim = event.classes[player.class];
  local rol_lim = event.roles[player.role];

  -- Check if exceeding class limit
  if(clas_lim.max and clas_lim.sort_stamp_count > clas_lim.max) -- Class limit reached
  then
    GEM3_ChatDebug(GEM3_DEBUG_SORT,"Plugin Stamp: Class limit reached for "..player.name..", check minimas ("..clas_lim.sort_stamp_count.." > "..tostring(clas_lim.min)..")");
    if(clas_lim.min and clas_lim.sort_stamp_count <= clas_lim.min)
    then
      GEM3_ChatDebug(GEM3_DEBUG_SORT,"Plugin Stamp: Class minima fails if I remove this player, keeping him as titular");
      return true;
    end
    return false;
  end
  return true; -- Check is OK
end

local function _GEM3_SORT_CheckRoleLimits(event,player,count)
  local clas_lim = event.classes[player.class];
  local rol_lim = event.roles[player.role];

  -- Check if exceeding role limit
  if(rol_lim and rol_lim.max and rol_lim.sort_stamp_count > rol_lim.max) -- Role limit reached
  then
    GEM3_ChatDebug(GEM3_DEBUG_SORT,"Plugin Stamp: Role limit reached for "..player.name..", check minimas ("..rol_lim.sort_stamp_count.." > "..tostring(rol_lim.min)..")");
    if(rol_lim.min and rol_lim.sort_stamp_count <= rol_lim.min)
    then
      GEM3_ChatDebug(GEM3_DEBUG_SORT,"Plugin Stamp: Role minima fails if I remove this player, keeping him as titular");
      return true;
    end
    return false;
  elseif(not _GEM3_SORT_DpsRoomForMe(event,player))
  then
    GEM3_ChatDebug(GEM3_DEBUG_SORT,"Plugin Stamp: DPS limit reached for "..player.name);
    return false;
  end
  return true; -- Check is OK
end

local function _GEM3_SORT_CheckMaxs(event,player,count)
  local clas_lim = event.classes[player.class];
  local rol_lim = event.roles[player.role];

  -- Check if exceeding class limit
  if(clas_lim.max and clas_lim.sort_stamp_count > clas_lim.max) -- Class limit reached
  then
    return false;
  elseif(rol_lim and rol_lim.max and rol_lim.sort_stamp_count > rol_lim.max) -- Role limit reached
  then
    return false;
  elseif(not _GEM3_SORT_DpsRoomForMe(event,player))
  then
    return false;
  elseif(count > event.max_count) -- Global limit reached
  then
    return false;
  end
  return true; -- Check is OK
end

local _GEM3_SORT_LeadersGuild;
local function _GEM3_SORT_SortList(tab1,tab2)
  if(tab1.lead_force_queue == GEM3_SUB_FORCE_QUEUE_TITULAR) -- 1 forced titular
  then
    if(tab2.lead_force_queue ~= GEM3_SUB_FORCE_QUEUE_TITULAR) -- but 2 not forced
    then
      return true; -- Sort 1 before
    end
  elseif(tab2.lead_force_queue == GEM3_SUB_FORCE_QUEUE_TITULAR) -- 2 forced titular (but not 1, otherwise it would have been trapped before)
  then
    return false; -- Sort 2 before
  end
  local inf1 = GEM3_PLAY_GetPlayerInfos(tab1.name);
  local inf2 = GEM3_PLAY_GetPlayerInfos(tab2.name);
  if(inf1 == nil or inf2 == nil) -- No infos?
  then
    if(inf1 and inf1.guild == _GEM3_SORT_LeadersGuild) -- Same guild than leader?
    then
      return true; -- Sort 1 first
    end
    if(inf2 and inf2.guild == _GEM3_SORT_LeadersGuild) -- Same guild than leader?
    then
      return false; -- Sort 2 first
    end
    return tab1.main_time < tab2.main_time; -- No info at all, sort by stamp
  end
  if(inf1.guild ~= _GEM3_SORT_LeadersGuild or inf2.guild ~= _GEM3_SORT_LeadersGuild) -- Not leader's guild
  then
    if(inf1.guild == _GEM3_SORT_LeadersGuild) -- Same guild than leader?
    then
      return true; -- Sort 1 first
    end
    if(inf2.guild == _GEM3_SORT_LeadersGuild) -- Same guild than leader?
    then
      return false; -- Sort 2 first
    end
    return tab1.main_time < tab2.main_time; -- No info at all, sort by stamp
  end
  -- Both are from the leader's guild, sort by rank
  if(inf1.grank_idx ~= inf2.grank_idx)
  then
    return inf1.grank_idx < inf2.grank_idx;
  end
  -- Same rank, sort by stamp
  return tab1.main_time < tab2.main_time;
end

local function _GEM3_SORT_SetTitular(event,tits,tab)
  local clas_lim = event.classes[tab.class];
  local rol_lim = event.roles[tab.role];

  clas_lim.sort_stamp_count = clas_lim.sort_stamp_count + 1; -- Increase class limit
  if(rol_lim)
  then
    rol_lim.sort_stamp_count = rol_lim.sort_stamp_count + 1; -- Increase role limit
  end
  tinsert(tits,tab);
end

local function _GEM3_SORT_SetSubstitute(event,subs,tab)
  tinsert(subs,tab);
end

local function _GEM3_SORT_MoveTitularToSubstitute(event,tits,subs,tab,pos)
  local clas_lim = event.classes[tab.class];
  local rol_lim = event.roles[tab.role];

  clas_lim.sort_stamp_count = clas_lim.sort_stamp_count - 1; -- Decrease class limit
  if(rol_lim)
  then
    rol_lim.sort_stamp_count = rol_lim.sort_stamp_count - 1; -- Decrease role limit
  end

  tinsert(subs,tab);
  tremove(tits,pos);
end

local function _GEM3_SORT_MoveTitularToRefillQueue(event,tits,refills,tab,pos)
  local clas_lim = event.classes[tab.class];
  local rol_lim = event.roles[tab.role];

  clas_lim.sort_stamp_count = clas_lim.sort_stamp_count - 1; -- Decrease class limit
  if(rol_lim)
  then
    rol_lim.sort_stamp_count = rol_lim.sort_stamp_count - 1; -- Decrease role limit
  end

  tinsert(refills,1,tab);
  tremove(tits,pos);
end

local function _GEM3_SORT_Sort(event)
  local tits = GA_GetTable();
  local subs = GA_GetTable();
  local repl = GA_GetTable();
  local rejs = GA_GetTable();
  local refills = GA_GetTable();

  GEM3_ChatDebug(GEM3_DEBUG_SORT,"Plugin Stamp: ## Entry");
  
  -- Reset some internal data
  for n,lim in pairs(event.classes)
  do
    lim.sort_stamp_count = 0;
  end
  for n,lim in pairs(event.roles)
  do
    lim.sort_stamp_count = 0;
  end

  -- Create list of all players eligible for Titular and Substitutes queues
  for name,tab in pairs(event.players)
  do
    GEM3_ChatDebug(GEM3_DEBUG_SORT,"Plugin Stamp: Processing player "..name..": LeadForced="..tostring(tab.lead_force_queue).." Forced="..tostring(tab.force_queue));
    if(tab.lead_force_queue == GEM3_SUB_FORCE_QUEUE_TITULAR and tab.force_queue ~= GEM3_SUB_FORCE_QUEUE_NOT_COMING) -- Forced Tit by leader, and still coming
    then
      _GEM3_SORT_SetTitular(event,tits,tab);
      GEM3_ChatDebug(GEM3_DEBUG_SORT,"Plugin Stamp: Forced titular by leader");
    elseif(tab.lead_force_queue == GEM3_SUB_FORCE_QUEUE_SUBSTITUTE and tab.force_queue ~= GEM3_SUB_FORCE_QUEUE_NOT_COMING) -- Forced Sub by leader, and still coming
    then
      _GEM3_SORT_SetSubstitute(event,subs,tab);
      GEM3_ChatDebug(GEM3_DEBUG_SORT,"Plugin Stamp: Forced substiture by leader");
    elseif(tab.lead_force_queue == GEM3_SUB_FORCE_QUEUE_REPLACEMENT and tab.force_queue ~= GEM3_SUB_FORCE_QUEUE_NOT_COMING) -- Forced Repl by leader, and still coming
    then
      tinsert(repl,tab);
      GEM3_ChatDebug(GEM3_DEBUG_SORT,"Plugin Stamp: Forced replacement by leader");
    elseif(tab.lead_force_queue == GEM3_SUB_FORCE_QUEUE_NOT_COMING) -- Forced NC by leader
    then
      tinsert(rejs,tab);
      GEM3_ChatDebug(GEM3_DEBUG_SORT,"Plugin Stamp: Forced reject by leader");
    elseif(tab.lead_force_queue == nil)
    then
      GEM3_ChatWarning("Plugin Stamp: tab.lead_force_queue is nil for "..name);
    else -- Use player's queue (not forced by leader, or player choose NOT_COMING)
      if(tab.force_queue == GEM3_SUB_FORCE_QUEUE_TITULAR) -- Player wants to be titular
      then
        GEM3_ChatDebug(GEM3_DEBUG_SORT,"Plugin Stamp: Player wants any available queue");
        _GEM3_SORT_SetTitular(event,tits,tab);
      elseif(tab.force_queue == GEM3_SUB_FORCE_QUEUE_SUBSTITUTE) -- Player want to be forced into Sub
      then
        _GEM3_SORT_SetSubstitute(event,subs,tab);
        GEM3_ChatDebug(GEM3_DEBUG_SORT,"Plugin Stamp: Forced substitute by player");
      elseif(tab.force_queue == GEM3_SUB_FORCE_QUEUE_REPLACEMENT) -- Player want to be forced into Repl
      then
        tinsert(repl,tab);
        GEM3_ChatDebug(GEM3_DEBUG_SORT,"Plugin Stamp: Forced replacement by player");
      elseif(tab.force_queue == GEM3_SUB_FORCE_QUEUE_NOT_COMING) -- Player don't want to come
      then
        tinsert(rejs,tab);
        GEM3_ChatDebug(GEM3_DEBUG_SORT,"Plugin Stamp: Forced reject by player");
      end
    end
  end

  -- Sort tits by stamp
  local leadinf = GEM3_PLAY_GetPlayerInfos(event.leader,event.channel);
  _GEM3_SORT_LeadersGuild = leadinf.guild;
  tsort(tits,_GEM3_SORT_SortList);
  
  GEM3_ChatDebug(GEM3_DEBUG_SORT,"Plugin Stamp: ## Pass 1 - Role&Class");
  local i = #tits; -- Start from last
  -- Move limit exceeding Tits intos Subs
  while(i ~= 0)
  do
    local tab = tits[i];
    GEM3_ChatDebug(GEM3_DEBUG_SORT,"Plugin Stamp: Check limits for player "..tab.name.." (i="..i..", count="..#tits..")");
    if(not _GEM3_SORT_CheckRoleClassLimits(event,tab,#tits))
    then
      _GEM3_SORT_MoveTitularToSubstitute(event,tits,subs,tab,i);
    elseif(not _GEM3_SORT_CheckGlobalLimit(event,tab,#tits)) -- Global limit reached
    then
      _GEM3_SORT_MoveTitularToRefillQueue(event,tits,refills,tab,i);
    end
    i = i - 1; -- Next!
  end
  
  GEM3_ChatDebug(GEM3_DEBUG_SORT,"Plugin Stamp: ## Pass 2 - Class");
  local i = #tits; -- Start from last
  -- Move limit exceeding Tits intos Subs
  while(i ~= 0)
  do
    local tab = tits[i];
    GEM3_ChatDebug(GEM3_DEBUG_SORT,"Plugin Stamp: Check single limits for player "..tab.name.." (i="..i..", count="..#tits..")");
    if(not _GEM3_SORT_CheckClassLimits(event,tab,#tits))
    then
      _GEM3_SORT_MoveTitularToSubstitute(event,tits,subs,tab,i);
    elseif(not _GEM3_SORT_CheckGlobalLimit(event,tab,#tits)) -- Global limit reached
    then
      _GEM3_SORT_MoveTitularToRefillQueue(event,tits,refills,tab,i);
    end
    i = i - 1; -- Next!
  end
  
  GEM3_ChatDebug(GEM3_DEBUG_SORT,"Plugin Stamp: ## Pass 3 - Role");
  local i = #tits; -- Start from last
  -- Move limit exceeding Tits intos Subs
  while(i ~= 0)
  do
    local tab = tits[i];
    GEM3_ChatDebug(GEM3_DEBUG_SORT,"Plugin Stamp: Check single limits for player "..tab.name.." (i="..i..", count="..#tits..")");
    if(not _GEM3_SORT_CheckRoleLimits(event,tab,#tits))
    then
      _GEM3_SORT_MoveTitularToSubstitute(event,tits,subs,tab,i);
    elseif(not _GEM3_SORT_CheckGlobalLimit(event,tab,#tits)) -- Global limit reached
    then
      _GEM3_SORT_MoveTitularToRefillQueue(event,tits,refills,tab,i);
    end
    i = i - 1; -- Next!
  end
  
  GEM3_ChatDebug(GEM3_DEBUG_SORT,"Plugin Stamp: ## Refills");
  -- Refill the titulars if too many were removed
  if(#tits < event.max_count and #refills ~= 0)
  then
    GEM3_ChatDebug(GEM3_DEBUG_SORT,"Plugin Stamp: Too many players removed from titulars, must refill queue ("..#tits.."/"..event.max_count..")");
    while(#tits < event.max_count)
    do
      if(#refills == 0)
      then
        break;
      end
      _GEM3_SORT_SetTitular(event,tits,refills[1]);
      GEM3_ChatDebug(GEM3_DEBUG_SORT,"Plugin Stamp: Refilling with "..refills[1].name.." ("..#tits.."/"..event.max_count..")");
      tremove(refills,1);
    end
  end

  -- Move refills into subs
  for _,tab in ipairs(refills)
  do
    tinsert(subs,tab);
  end

  GEM3_ChatDebug(GEM3_DEBUG_SORT,"Plugin Stamp: ## Summary");
  -- Sort other queues by stamp
  tsort(subs,_GEM3_SORT_SortList);
  tsort(repl,_GEM3_SORT_SortList);
  tsort(rejs,_GEM3_SORT_SortList);
  
  local consistency = true;
  -- Set queue pos for all queues
  for i,tab in ipairs(tits)
  do
    tab.current_queue = GEM3_SUB_STATE_TITULAR;
    tab.queue_pos = i;
    GEM3_ChatDebug(GEM3_DEBUG_SORT,"Plugin Stamp: Titular #"..i.." is "..tab.name);
    if(not _GEM3_SORT_CheckMaxs(event,tab,#tits))
    then
      GEM3_ChatDebug(GEM3_DEBUG_SORT,"Plugin Stamp: Min/Max consistency failure detected");
      consistency = false;
    end
  end
  for i,tab in ipairs(subs)
  do
    tab.current_queue = GEM3_SUB_STATE_SUBSTITUTE;
    tab.queue_pos = i;
    GEM3_ChatDebug(GEM3_DEBUG_SORT,"Plugin Stamp: Substitute #"..i.." is "..tab.name);
  end
  for i,tab in ipairs(repl)
  do
    tab.current_queue = GEM3_SUB_STATE_REPLACEMENT;
    tab.queue_pos = i;
    GEM3_ChatDebug(GEM3_DEBUG_SORT,"Plugin Stamp: Replacement #"..i.." is "..tab.name);
  end
  for i,tab in ipairs(rejs)
  do
    tab.current_queue = GEM3_SUB_STATE_NOT_COMING;
    tab.queue_pos = i;
    GEM3_ChatDebug(GEM3_DEBUG_SORT,"Plugin Stamp: NotComing #"..i.." is "..tab.name);
  end

  -- Release tables
  GA_ReleaseTable(tits);
  GA_ReleaseTable(subs);
  GA_ReleaseTable(repl);
  GA_ReleaseTable(rejs);
  GA_ReleaseTable(refills);

  if(not consistency)
  then
    -- GUI Callback
    GEM3_TriggerCallback("OnSortingInconsistency",event);
  end
  GEM3_ChatDebug(GEM3_DEBUG_SORT,"Plugin Stamp: ## Exit");
end


--------------- Plugin structure ---------------

local GEM3_SORT_Stamp = {
  --[[
    Name parameter. [MANDATORY]
     Displayed name of the sorting plugin.
  ]]
  Name = GEM3_SORT_NAME;
  
  --[[
    SortType parameter. [MANDATORY]
     Internal code for the sorting type (must be unique).
     Please request a unique SortType value to Kiki, if you plan on creating a sorting plugin yourself.
  ]]
  SortType = 3;
  
  --[[
    Subscribers sorting function. [MANDATORY]
     Sorts all passed players in two lists.
     Params :
      - event : Event (STRUCT)
  ]]
  Sort = function(event)
    _GEM3_SORT_Sort(event);
  end;
  
  --[[
    Configure function. [OPTIONAL]
     Configures the plugin.
  ]]
  --Configure = function()
  --end;

  --[[
    Help parameter. [OPTIONAL]
     Help string displayed when you mouse over the sort type.
  ]]
  Help = GEM3_SORT_HELP;
};

-- Register Plugin
GEM3_SORT_RegisterPlugin(GEM3_SORT_Stamp);

