﻿function MovixLib_OnReceiveDrag(...)
 local self = select(1, ...);
 local frame = self.parent; 
 if not frame.dragingCol then return; end
 local i = 0;
 local diff = 0
 if self.hid > frame.dragingCol.hid then --- if is dragged after
   for colind, column in pairs(frame.Headers) do
    i = i +1;
    if i> frame.dragingCol.hid and i <= self.hid then
        MovixLib_GridSetNewColPos(frame, column, column.pos, column.pos - 1)
    end
    if i == frame.dragingCol.hid then
        MovixLib_GridSetNewColPos(frame, column, column.pos, self.hid)
    end
   end
 else --- if is dragged prior
   for colind, column in pairs(frame.Headers) do
    i = i +1;
    if i>= self.hid and i < frame.dragingCol.hid then
      MovixLib_GridSetNewColPos(frame, column, column.pos, column.pos + 1)
    end
    if i == frame.dragingCol.hid then
     MovixLib_GridSetNewColPos(frame, column, column.pos, self.hid)
    end
   end
 end

 table.sort(frame.Headers, mycolsort);
 MovixLib_RefreshGrid(frame, true)
 frame.dragingCol = nil;
end

function  MovixLib_ColDrag(...)
   local self = select(1, ...);
   local frame = self.parent:GetParent():GetParent();
   local x, y = GetCursorPosition();
   local newx = self.x - (self.tl - x);
   self:SetPoint("TOPLEFT", self.rt, "TOPLEFT", newx, self.y);
end


function MovixLib_StartColDrag(...)
   local self = select(1, ...);
   local frame = self.parent;
   frame.dragingCol = self;
   self.oldpos = frame.Headers[self.hid].pos;
   local p, rt, rp, x, y = self:GetPoint(1);
   self.rt=rt;
   self.x =x;
   self.y=y;
   local x, y = GetCursorPosition();
   self.tl = x;
   self:SetScript("OnUpdate", MovixLib_ColDrag);
   self.framestrata=self:GetFrameStrata();
   self:SetFrameStrata("BACKGROUND");
end


function MovixLib_StopColDrag(...)
   local self = select(1, ...);
   local frame = self.parent;
   self:SetScript("OnUpdate", nil);
   self:SetFrameStrata(self.framestrata);
   self:SetUserPlaced(false);
   MovixLib_RefreshGrid(frame, true)
end






function  MovixLib_ResizeCol(...)
   local self = select(1, ...);
   local frame = self.parent;
   local x, y = GetCursorPosition();
   local neww = self.oldw - (self.tl - x);
   if neww < Movixlib_GetGridParam(frame.gridid, "Mincolwidth") then neww = Movixlib_GetGridParam(frame.gridid, "Mincolwidth") end
   frame.Headers[self.hid].size = neww ;
   MovixLib_RefreshGrid(frame)
end


function MovixLib_StartResizeCol(...)
   local self = select(1, ...);
   local frame = self.parent;
   if not frame then return end
   self.oldw = frame.Headers[self.hid].size;
   local x, y = GetCursorPosition();
   self.tl = x;
   self:StartMoving();
   self:SetScript("OnUpdate", MovixLib_ResizeCol);
end


function MovixLib_StopResizeCol(...)
   local self = select(1, ...);
   local frame = self.parent;
   self:SetScript("OnUpdate", nil);
   self:StopMovingOrSizing();
   self:SetUserPlaced(false);
   MovixLib_RefreshGrid(frame, true)
end



function  MovixLib_ResizeHeader(...)
   local self = select(1, ...);
   local frame = self.parent;
   local x, y = GetCursorPosition();
   local newh = self.oldh + (self.tl - y);
   if newh < 15 then newh = 15; end
   if newh > frame:GetHeight()/2 then newh = frame:GetHeight()/2; end
   frame.GridParams.headerheight = newh ;
   MovixLib_RefreshGrid(frame)
end


function MovixLib_StartResizeHeader(...)
   local self = select(1, ...);
   local frame = self.parent;
   self.oldh = Movixlib_GetGridParam(frame.gridid, "headerheight")
   local x, y = GetCursorPosition();
   self.tl = y;
   self:StartMoving();
   self:SetScript("OnUpdate", MovixLib_ResizeHeader);
end


function MovixLib_StopResizeHeader(...)
   local self = select(1, ...);
   local frame = self.parent;
   self:SetScript("OnUpdate", nil);
   self:StopMovingOrSizing();
   self:SetUserPlaced(false);
   MovixLib_RefreshGrid(frame, true)
end

function MovixLib_OnEnterColSep()
 SetCursor("CAST_CURSOR")
end

function MovixLib_OnLeaveColSep()
 SetCursor(nil)
end

function MovixLib_BuildColumnHeaderTooltip(gridid, lbl, tooltip)
 GameTooltip:ClearLines();
 -- Tooltip Title
 if lbl == nil then
    lbl = ""
 end
 if string.sub(lbl,1,2) == "#_" then
    lbl = string.sub(lbl,3)
    lbl = MovixLib_GetLocale(lbl,gridid)
    GameTooltip:SetText(lbl, 1,1,1,1);
 elseif string.sub(lbl,1,4) == "#GS_" then
    local foo = loadstring("return "..string.sub(lbl,5))
    lbl = foo()
    GameTooltip:SetText(lbl, 1,1,1,1);
 else
    lbl = MovixLib_GetLocale(lbl,gridid)
    GameTooltip:SetText(lbl, 1,1,1,1);
 end
 
 -- tooltip text
 tooltip = MovixLib_GetLocale(tooltip, gridid)
 local lines = {strsplit("\\", tooltip or "")}
 for _, line in pairs(lines) do
    if line ~= "" then 
     GameTooltip:AddLine(line, 1,1,0,1); 
    end
 end
 return lbl, tooltip
end

function MovixLib_OnEnterGridHeader(self) 
 local p = self.parent
 local col = p.Headers[self.hid]
 GameTooltip:SetOwner(self, "ANCHOR_LEFT"); 
 MovixLib_BuildColumnHeaderTooltip(p.gridid, col.lbl, col.tooltip)
 GameTooltip:Show();
 --GameTooltip:SetFrameStrata("TOOLTIP");
 --GameTooltip:SetFrameLevel(255);
 --GameTooltip:Raise();
 
 -- display sorting arrow if needed
 getglobal(self:GetName().."sorttext"):Show()
end

function MovixLib_OnLeaveGridHeader(self)
 GameTooltip:Hide();
 getglobal(self:GetName().."sorttext"):Hide()
end


function MovixLib_OnClickLine(...)
 local clickfoo
 local self, button = select(1, ...);
 local p = self.parent
 local row = self.rowindex
 local col = nil
 local clickfoo = nil
 local oldhlt = nil
 local hlt = nil

 if button == "LeftButton" then
  if IsControlKeyDown() then
      if p.SelectedRow then
          if not p.SelectedRow[p.data[row].MovixLibID] then
            p.SelectedRow[p.data[row].MovixLibID] = self
            getglobal(self:GetName().."HL"):Show()
            p.data[row].MovixLibSelected = true
          else
            p.SelectedRow[p.data[row].MovixLibID] = nil
            getglobal(self:GetName().."HL"):Hide()
            p.data[row].MovixLibSelected = false
          end
      else
          p.SelectedRow = {}
          p.SelectedRow[p.data[row].MovixLibID] = self
          getglobal(self:GetName().."HL"):Show()
          p.data[row].MovixLibSelected = false
      end
  else
    local allreadyselected = false
    if p.SelectedRow then
        for key, line in pairs (p.SelectedRow) do
            if key == p.data[row].MovixLibID then
                allreadyselected = true
            end
            p.SelectedRow[p.data[row].MovixLibID] = nil
            p.data[row].MovixLibSelected = false
            getglobal(line:GetName().."HL"):Hide()
        end
    end
    if not allreadyselected then
        p.SelectedRow = {}
        p.SelectedRow[p.data[row].MovixLibID] = self
        getglobal(self:GetName().."HL"):Show()
        p.data[row].MovixLibSelected = true
    end
  end
  local i = 0
  for ind, sel in pairs (p.SelectedRow) do
   i = i +1
  end
  if i > 1 then
    if p.Headers[1].val ~= "MovixLibSelected" then 
        local colparams = MovixLib_NewCellOptions("MovixLibSelected", "")
        MovixLib_SetCellOption(colparams, "type", "boolean")
        MovixLib_SetCellOption(colparams, "lbl", "")
        MovixLib_SetCellOption(colparams, "size", 12)
        MovixLib_SetCellOption(colparams, "valminh", 10)
        MovixLib_SetCellOption(colparams, "valminw", 10)
        MovixLib_SetCellOption(colparams, "valmaxw", 10)
        MovixLib_SetCellOption(colparams, "valmaxh", 10)
        
        if p.GridParams.sortcol then
            for i, cs in ipairs(p.GridParams.sortcol) do
                p.GridParams.sortcol[i] = (math.abs(cs)+1)*(cs/math.abs(cs))
            end
            table.insert(p.GridParams.sortcol, 1, 1)
        else
            p.GridParams.sortcol = {}
            table.insert(p.GridParams.sortcol, 1, 1)
        end
        MovixLib_AddCol(p, nil, "MovixLibSelected", colparams)
    end
    MovixLib_FillData(p)
    MovixLib_RefreshGrid(p)
  else
    if p.Headers[1].val == "MovixLibSelected" then 
        table.remove(p.Headers, 1)
        if p.GridParams.sortcol then
            for i, cs in ipairs(p.GridParams.sortcol) do
                p.GridParams.sortcol[i] = (math.abs(cs)-1)*(cs/math.abs(cs))
            end
            table.remove(p.GridParams.sortcol, 1)
        end
        if #p.GridParams.sortcol == 0 then p.GridParams.sortcol = nil end
        MovixLib_Grid_ResetColumnOrder(p.Headers)
        MovixLib_FillData(p)
        MovixLib_RefreshGrid(p, true)
    end
  end

  clickfoo = Movixlib_GetGridParam(p.gridid, "OnClickFoo")
  Movixlib_Grid_ExecuteCustomFoo(p, row, col,clickfoo)
 elseif button == "RightButton" then
  Movixlib_LineRightClick(...)
 end
end

function MovixLib_OnDoubleClickLine(...)
    local self, button = select(1, ...);
    local p = self.parent
    local row = self.rowindex
    clickfoo = Movixlib_GetGridParam(p.gridid, "OnDclickFoo")
    Movixlib_Grid_ExecuteCustomFoo(p, row, col,clickfoo)
end

function Movixlib_LineRightClick(...)
    local self, button = select(1, ...);
    local p = self.parent
    local row = self.rowindex
    MovixLib_DDTable = MovixLib_ReleaseTable(MovixLib_DDTable, 10);
    local clickfoo = Movixlib_GetGridParam(p.gridid, "OnRclickFoo")
    Movixlib_Grid_ExecuteCustomFoo(p, row, col,clickfoo)
    
    if MovixLib_DDTable and #MovixLib_DDTable > 0 then
        MovixLib_DDMenuAddLine({text = " ", disabled = true})
    end
    MovixLib_GetGridMenu(p)    
    MovixLib_SetDDMenu(self)
end


function MovixLib_OnEnterLine (...)
 local self = select(1, ...);
 local p = self.parent
 local row = self.rowindex
 local col = nil
 p.mohlt_t:Show()
 p.mohlt_b:Show()
 p.mohlt_t:SetParent(self)
 p.mohlt_b:SetParent(self)
 p.mohlt_t:ClearAllPoints()
 p.mohlt_b:ClearAllPoints()
 p.mohlt_t:SetPoint("TOPLEFT",0,0)
 p.mohlt_b:SetPoint("BOTTOMLEFT",0,0)
 local foo = Movixlib_GetGridParam(p.gridid, "OnEnterFoo")
 Movixlib_Grid_ExecuteCustomFoo(p, row, col,foo)
end
 
function MovixLib_OnLeaveLine (...)
local self = select(1, ...);
 local p = self.parent
 if not p then return end
 local row = self.rowindex
 local col = nil
 p.mohlt_t:Hide()
 p.mohlt_b:Hide()
 local foo = Movixlib_GetGridParam(p.gridid, "OnLeaveFoo")
 Movixlib_Grid_ExecuteCustomFoo(p, row, col,foo)
end


function MovixLib_Grid_ExecuteFormFoo(p, row, colframe, foo, action, scale)
 local retval = nil
 if foo then
         local foo = loadstring([[
                            return
                            function (p, row, colframe, action, scale)
                                return ]] .. foo .. [[(p, row, colframe, action, scale)
                            end
                         ]]
                        )
    func = foo();
    retval = func(p, row, colframe, action, scale)
 end
 return retval
end

function Movixlib_Grid_ExecuteCustomFoo(p, row, col, foo)
 local retval = nil
 if foo then
         local foo = loadstring([[
                            return
                            function (p, row, col)
                                return ]] .. foo .. [[(p, row, col)
                            end
                         ]]
                        )
    func = foo();
    retval = func(p, row, col)
 end
 return retval
end


function Movixlib_Grid_ExecuteErrorfoo(foo)
         local f = loadstring([[
                            return
                            function ()
                                return ]] .. foo .. [[()
                            end
                         ]]
                        )
    func = f();
    return func()
end

function Movixlib_Grid_ExecuteClickfoo(self)
    local foo = self.clickfoo
    local val = self.val
    local f = loadstring([[
                            return
                            function (val)
                                return ]] .. foo .. [[(val)
                            end
                         ]]
                        )
    func = f();
    func(val)
end